from universe.vectorized.core import Env, Wrapper, ObservationWrapper, ActionWrapper, RewardWrapper
from universe.vectorized.multiprocessing_env import MultiprocessingEnv
from universe.vectorized.vectorize_filter import Filter, VectorizeFilter



import logging
import os
import re
import six.moves.urllib.parse as urlparse

from universe import error, utils
from universe.remotes import remote

logger = logging.getLogger(__name__)

class HardcodedAddresses(object):
    @classmethod
    def build(cls, remotes, **kwargs):
        parsed = urlparse.urlparse(remotes)
        if parsed.scheme != 'vnc':
            raise error.Error('HardcodedAddresses must be initialized with a string starting with vnc://: {}'.format(remotes))

        addresses = parsed.netloc.split(',')
        query = urlparse.parse_qs(parsed.query)
        # We could support per-backend passwords, but no need for it
        # right now.
        password = query.get('password', [utils.default_password()])[0]
        vnc_addresses, rewarder_addresses = parse_remotes(addresses)
        res = cls(vnc_addresses, rewarder_addresses, vnc_password=password, rewarder_password=password, **kwargs)
        return res, res.available_n

    def __init__(self, vnc_addresses, rewarder_addresses, vnc_password, rewarder_password, start_timeout=None):
        if vnc_addresses is not None:
            self.available_n = len(vnc_addresses)
        elif rewarder_addresses is not None:
            self.available_n = len(rewarder_addresses)
        else:
            assert False

        self.supports_reconnect = False
        self.connect_vnc = vnc_addresses is not None
        self.connect_rewarder = rewarder_addresses is not None

        self.vnc_addresses = vnc_addresses
        self.vnc_password = vnc_password
        self.rewarder_addresses = rewarder_addresses
        self.rewarder_password = rewarder_password
        if start_timeout is None:
            start_timeout = 2 * self.available_n + 5
        self.start_timeout = start_timeout

        self._popped = False

    def pop(self, n=None):
        if self._popped:
            assert n is None
            return []
        self._popped = True

        remotes = []
        for i in range(self.available_n):
            if self.vnc_addresses is not None:
                vnc_address = self.vnc_addresses[i]
            else:
                vnc_address = None

            if self.rewarder_addresses is not None:
                rewarder_address = self.rewarder_addresses[i]
            else:
                rewarder_address = None

            name = self._handles[i]
            env = remote.Remote(
                handle=self._handles[i],
                vnc_address=vnc_address,
                vnc_password=self.vnc_password,
                rewarder_address=rewarder_address,
                rewarder_password=self.rewarder_password,
            )
            remotes.append(env)
        return remotes

    def allocate(self, handles, initial=False, params={}):
        if len(handles) > self.available_n:
            raise error.Error('Requested {} handles, but only have {} envs'.format(len(handles), self.available_n))
        self.n = len(handles)
        self._handles = handles

    def close(self):
        pass

def parse_remotes(remotes):
    # Parse a list of remotes of the form:
    #
    # address:vnc_port+rewarder_port (e.g. localhost:5900+15900)
    #
    # either vnc_port or rewarder_port can be omitted, but not both

    all_vnc = None
    all_rewarder = None

    vnc_addresses = []
    rewarder_addresses = []

    for remote in remotes:
        # Parse off +, then :
        if '+' in remote:
            if all_vnc == False:
                raise error.Error('Either all or no remotes must have rewarders: {}'.format(remotes))
            all_vnc = True

            remote, rewarder_port = remote.split('+')
            rewarder_port = int(rewarder_port)
        else:
            if all_vnc == True:
                raise error.Error('Either all or no remotes must have rewarders: {}'.format(remotes))
            all_vnc = False

            rewarder_port = None

        if ':' in remote:
            if all_rewarder == False:
                raise error.Error('Either all or no remotes must have a VNC port: {}'.format(remotes))
            all_rewarder = True

            remote, vnc_port = remote.split(':')
            vnc_port = int(vnc_port)
        else:
            if all_rewarder == True:
                raise error.Error('Either all or no remotes must have a VNC port: {}'.format(remotes))
            all_rewarder = False

            vnc_port = None
            all_rewarder = False

        host = remote

        if rewarder_port is not None:
            rewarder_address = '{}:{}'.format(host, rewarder_port)
            rewarder_addresses.append(rewarder_address)

        if vnc_port is not None:
            vnc_address = '{}:{}'.format(host, vnc_port)
            vnc_addresses.append(vnc_address)

    if not all_vnc and not all_rewarder:
        raise error.Error('You must provide either rewarder or a VNC port: {}'.format(remotes))

    if not vnc_addresses:
        vnc_addresses = None
    if not rewarder_addresses:
        rewarder_addresses = None
    return vnc_addresses, rewarder_addresses



import logging
import numpy as np
from universe import vectorized
from universe.wrappers import render

logger = logging.getLogger(__name__)

def WrappedMultiprocessingEnv(env_id):
    return render.Render(EpisodeID(vectorized.MultiprocessingEnv(env_id)))

class RemoveNones(vectorized.Wrapper):
    """The vectorized environment will return None for any indexes that
    have already exceeded their episode count (not to be confused with
    the Nones returned by resetting environments in the real-time
    case). For convenience, we instead return a plausible observation
    in each such slot.
    """
    def __init__(self, env):
        super(RemoveNones, self).__init__(env)
        self.plausible_observation = None

    def _reset(self):
        observation_n = self.env.reset()
        self.plausible_observation = observation_n[0]
        return observation_n

    def _step(self, action_n):
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        observation_n = [ob if ob is not None else self.plausible_observation for ob in observation_n]
        return observation_n, reward_n, done_n, info

class EpisodeID(vectorized.Wrapper):
    """
For each episode, return its id, and also return the total number of contiguous 
episodes that are now done. 
"""
    def _configure(self, episode_limit=None, **kwargs):
        super(EpisodeID, self)._configure(**kwargs)
        assert self.metadata.get('runtime.vectorized')
        self.episode_limit = episode_limit
        self._clear_state()

    def _clear_state(self):
        self.done_to = -1
        self.extra_done = set()
        self.episode_ids = list(range(self.n))

    def _set_done_to(self):
        while True:
            next_done_to = self.done_to + 1
            if next_done_to in self.extra_done:
                self.done_to = next_done_to
                self.extra_done.remove(next_done_to)
            else:
                break

    def _reset(self):
        self._clear_state()
        return self.env.reset()

    def _step(self, action_n):
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        # Pass along ID of potentially-done episode
        for i, info_i in enumerate(info['n']):
            info_i['vectorized.episode_id'] = self.episode_ids[i]

        done_i = np.argwhere(done_n).reshape(-1)
        if len(done_i):
            for i in done_i:
                self.extra_done.add(self.episode_ids[i])
                # Episode completed, so we bump its value
                self.episode_ids[i] += self.n
                if self.episode_limit is not None and self.episode_ids[i] > self.episode_limit:
                    logger.debug('Masking: index=%s episode_id=%s', i, self.episode_ids[i])
                    self.env.mask(i)
            self._set_done_to()

        # Pass along the number of contiguous episodes that are now done
        info['vectorized.done_to'] = self.done_to
        return observation_n, reward_n, done_n, info



# -*- coding: utf-8 -*-

import logging
import numpy as np
import six
import time

from universe import vectorized
from universe.utils import display

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

def stats(count):
    flat = [e for vals in count for e in vals]
    if len(flat) == 0:
        return '(empty)'
    s = '%0.1f±%0.1f' % (np.mean(flat), np.std(flat))
    if six.PY2:
        # There is not a great way to backport Unicode support to Python 2.
        # We don't use it much, anyway. Easier just not to try.
        s = s.replace('±', '+-')
    return s

class Logger(vectorized.Wrapper):
    def __init__(self, env):
        super(Logger, self).__init__(env)

    def _configure(self, print_frequency=5, **kwargs):
        self.print_frequency = print_frequency
        extra_logger.info('Running VNC environments with Logger set to print_frequency=%s. To change this, pass "print_frequency=k" or "print_frequency=None" to "env.configure".', self.print_frequency)
        super(Logger, self)._configure(**kwargs)
        self._clear_step_state()
        self.metadata['render.modes'] = self.env.metadata['render.modes']

        self._last_step_time = None

    def _clear_step_state(self):
        self.frames = 0
        self.last_print = time.time()
        # time between action being sent and processed
        self.action_lag_n = [[] for _ in range(self.n)]
        # time between observation being generated on the server and being passed to add_metadata
        self.observation_lag_n = [[] for _ in range(self.n)]
        # time between observation being passed to add_metadata and being returned to Logger
        self.processing_lag = []
        # time between observation being returned by Logger and then action being passed to Throttle
        self.thinking_lag = []

        self.vnc_updates_n = [[] for _ in range(self.n)]
        self.vnc_bytes_n = [[] for _ in range(self.n)]
        self.vnc_pixels_n = [[] for _ in range(self.n)]
        self.reward_count_n = [[] for _ in range(self.n)]
        self.reward_total_n = [[] for _ in range(self.n)]
        self.reward_lag_n = [[] for _ in range(self.n)]
        self.rewarder_message_lag_n = [[] for _ in range(self.n)]

    def _step(self, action_n):
        observation_n, reward_n, done_n, info = self.env.step(action_n)
        if self.print_frequency is None:
            return observation_n, reward_n, done_n, info

        last_step_time = self._last_step_time
        self._last_step_time = time.time()

        # Printing
        self.frames += 1
        delta = time.time() - self.last_print
        if delta > self.print_frequency:
            fps = self.frames/delta

            # Displayed independently
            # action_lag = ','.join([diagnostics.display_timestamps_pair_max(action_lag) for action_lag in self.action_lag_n])
            # observation_lag = ','.join([diagnostics.display_timestamps_pair_max(observation_lag) for observation_lag in self.observation_lag_n])

            flat = False

            # Smooshed together
            action_lag, action_data = display.compute_timestamps_pair_max(self.action_lag_n, flat=flat)
            observation_lag, observation_data = display.compute_timestamps_pair_max(self.observation_lag_n, flat=flat)
            processing_lag, processing_data = display.compute_timestamps_sigma(self.processing_lag)
            thinking_lag, thinking_data = display.compute_timestamps_sigma(self.thinking_lag)
            reward_count = [sum(r) / delta for r in self.reward_count_n]
            if flat and len(reward_count) > 0:
                reward_count = np.mean(reward_count)
            reward_total = [sum(r) / delta for r in self.reward_total_n]
            if flat and len(reward_total) > 0:
                reward_total = np.mean(reward_total)
            reward_lag, reward_data = display.compute_timestamps_pair_max(self.reward_lag_n, flat=flat)
            rewarder_message_lag, rewarder_message_data = display.compute_timestamps_pair_max(self.rewarder_message_lag_n, flat=flat)
            vnc_updates_count = [sum(v) / delta for v in self.vnc_updates_n]
            if flat and len(vnc_updates_count) > 0:
                vnc_updates_count = np.mean(vnc_updates_count)

            # Always aggregate these ones
            if len(self.vnc_bytes_n) > 0:
                vnc_bytes_count = np.sum(e for vnc_bytes in self.vnc_bytes_n for e in vnc_bytes) / delta
            else:
                vnc_bytes_count = None
            if len(self.vnc_pixels_n) > 0:
                vnc_pixels_count = np.sum(e for vnc_pixels in self.vnc_pixels_n for e in vnc_pixels) / delta
            else:
                vnc_pixels_count = None

            reward_stats = stats(self.reward_count_n)
            vnc_updates_stats = stats(self.vnc_updates_n)
            vnc_bytes_stats = stats(self.vnc_bytes_n)
            vnc_pixels_stats = stats(self.vnc_pixels_n)

            reaction_time = []
            for a, o in zip(action_data, observation_data):
                try:
                    value = thinking_data['mean'] + processing_data['mean'] + a['mean'] + o['mean']
                except KeyError:
                    reaction_time.append(None)
                else:
                    reaction_time.append(display.display_timestamp(value))

            log = []
            for key, spec, value in [
                    ('vnc_updates_ps', '%0.1f', vnc_updates_count),
                    ('n', '%s', self.n),
                    ('reaction_time', '%s', reaction_time),
                    ('observation_lag', '%s', observation_lag),
                    ('action_lag', '%s', action_lag),
                    ('processing_lag', '%s', processing_lag),
                    ('thinking_lag', '%s', thinking_lag),
                    ('reward_ps', '%0.1f', reward_count),
                    ('reward_total', '%0.1f', reward_total),
                    ('vnc_bytes_ps[total]', '%0.1f', vnc_bytes_count),
                    ('vnc_pixels_ps[total]', '%0.1f', vnc_pixels_count),
                    ('reward_lag', '%s', reward_lag),
                    ('rewarder_message_lag', '%s', rewarder_message_lag),
                    ('fps', '%0.2f', fps),
            ]:
                if value == None:
                    continue

                if isinstance(value, list):
                    value = ','.join(spec % v for v in value)
                else:
                    value = spec % value
                log.append('%s=%s' % (key, value))

            if not log:
                log.append('(empty)')

            if self.frames != 0:
                logger.info('Stats for the past %.2fs: %s', delta, ' '.join(log))
            self._clear_step_state()

        # These are properties of the step rather than any one index
        observation_available_at = info.get('throttle.observation.available_at')
        if observation_available_at is not None:
            # (approximate time that we're going to return -- i.e. now, assuming Logger is fast)
            # - (time that the observation was passed to add_metadata)
            self.processing_lag.append(self._last_step_time - observation_available_at)

        action_available_at = info.get('throttle.action.available_at')
        if action_available_at is not None and last_step_time is not None:
            # (time that the action was generated) - (approximate time that we last returned)
            self.thinking_lag.append(action_available_at - last_step_time)

        # Saving of lags
        for i, info_i in enumerate(info['n']):
            observation_lag = info_i.get('stats.gauges.diagnostics.lag.observation')
            if observation_lag is not None:
                self.observation_lag_n[i].append(observation_lag)

            action_lag = info_i.get('stats.gauges.diagnostics.lag.action')
            if action_lag is not None:
                self.action_lag_n[i].append(action_lag)

            reward_count = info_i.get('reward.count')
            if reward_count is not None:
                self.reward_count_n[i].append(reward_count)

            reward_total = reward_n[i]
            if reward_total is not None:
                self.reward_total_n[i].append(reward_total)

            assert 'vnc.updates.n' not in info, 'Looks like you are using an old go-vncdriver. Please update to >=0.4.0: pip install --ignore-installed --no-cache-dir go-vncdriver'

            vnc_updates = info_i.get('stats.vnc.updates.n')
            if vnc_updates is not None:
                self.vnc_updates_n[i].append(vnc_updates)

            vnc_bytes = info_i.get('stats.vnc.updates.bytes')
            if vnc_bytes is not None:
                self.vnc_bytes_n[i].append(vnc_bytes)

            vnc_pixels = info_i.get('stats.vnc.updates.pixels')
            if vnc_pixels is not None:
                self.vnc_pixels_n[i].append(vnc_pixels)

            reward_lag = info_i.get('stats.gauges.diagnostics.lag.reward')
            if reward_lag is not None:
                self.reward_lag_n[i].append(reward_lag)

            rewarder_message_lag = info_i.get('stats.gauges.diagnostics.lag.rewarder_message')
            if rewarder_message_lag is not None:
                self.rewarder_message_lag_n[i].append(rewarder_message_lag)

        return observation_n, reward_n, done_n, info



import json
import os
import struct

from universe import error

class InvalidFBSFileError(error.Error):
    pass

class FBSReader(object):
    def __init__(self, path):
        self.file = open(path, 'rb')
        version = self.file.read(12)
        if version != b'FBS 001.002\n':
            raise InvalidFBSFileError('Unrecognized FBS version: {}'.format(version))

        header = self.file.readline()
        pos = self.file.tell()
        self.file.seek(pos, os.SEEK_SET)

        header = json.loads(header.decode('utf-8'))
        self.start = header['start']

    def __iter__(self):
        return self

    def read_safe(self, size=None):
        """
        We currently close our fbs files by killing them, so sometimes they end
        up with bad data at the end. Close our reader if we expect `size` bytes
        and get fewer.

        This is a hack and should be removed when we cleanly close our
        connections in fbs_writer.

        https://github.com/openai/universe-envs/issues/41
        """
        bytes = self.file.read(size)
        if len(bytes) != size:
            # We unexpectedly got to the end of the file
            self.close()
            raise StopIteration
        return bytes

    def next(self):
        return self.__next__()

    def __next__(self):
        length_str = self.read_safe(4)
        if length_str == '':
            # Indicates a file with no trailer
            self.close()
            raise StopIteration
        (length,) = struct.unpack('!I', length_str)

        if length == 0:
            # Reached the end
            self.close()
            raise StopIteration()

        data = self.read_safe(length)
        timestamp_str = self.read_safe(4)
        (timestamp,) = struct.unpack('!I', timestamp_str)

        return data, self.start + timestamp/1000.

    def close(self):
        self.file.close()



import string

from universe import spaces
from universe.spaces import vnc_event, VNCActionSpace
from universe.spaces.vnc_event import KeyEvent, PointerEvent
from universe.envs import vnc_env
from universe.vncdriver import constants
import logging

logger = logging.getLogger()

SCREEN_DIM = (640, 480)

class StarCraftEnv(vnc_env.VNCEnv):
    def __init__(self):
        super(StarCraftEnv, self).__init__()
        self.action_space = VNCActionSpace(
            keys=['f2',  # Map positions
                  'f3',  # Map positions
                  'f4',  # Map positions
                  'spacebar',
                  'left',
                  'up',
                  'right',
                  'down'],
            screen_shape=SCREEN_DIM
        )
        self.safe_action_space = self.action_space

    # def _step(self, action_n):
    #     return super(StarCraftEnv, self)._step(
    #         (StarCraftEventFilter.filter(a) for a in action_n))


# class StarCraftEventFilter(object):
#     """
#     We only allow keyboard inputs used by StarCraft:
#     http://gamingweapons.com/image/steelseries/zboard-starcraft2-keyset/steelseries_zboard_starcraft2_keyset_02.jpg
#     """
#     _x_offset = 5  # Centered
#     _y_offset = 30  # Remove the chrome

#     @classmethod
#     def _safe_pointer_event(cls, event):
#         """Returns true if the click is in a place that will not break out of the box"""
#         height = SCREEN_DIM[0]
#         width = SCREEN_DIM[1]
#         margin = 5  # Never allow clicking within 5 pixels of the edge of the screen

#         unsafe_locations = [
#             (event.y < cls._y_offset + margin),  # At the top, where menu chrome is
#             (event.y > height + cls._y_offset - margin),  # Too far down
#             (event.x < cls._x_offset + margin),  # Too far left
#             (event.x > width + cls._x_offset - margin),  # Too far right
#             (410 < event.x < 510) and (370 < event.y < 450),  # Where the menu button is
#         ]
#         unsafe = any(unsafe_locations)
#         if unsafe:
#             logger.warning('skipping unsafe pointer event')
#         return not unsafe

#     @classmethod
#     def safe_event(cls, event):
#         if isinstance(event, PointerEvent):
#             return cls._safe_pointer_event(event)

#     @classmethod
#     def filter(cls, events):
#         return filter(cls.safe_event, events)



from universe.rewarder.rewarder_session import RewarderSession
from universe.rewarder.env_status import EnvStatus, compare_ids
from universe.rewarder.merge import merge_n, merge_infos, merge_reward_n, merge_observation_n
from universe.rewarder.reward_buffer import RewardBuffer



import collections
import json

import six
from gym import error


class UnregisteredRuntime(error.Unregistered):
    """Raised when the user requests a runtime from the registry that does
    not actually exist.
    """
    pass

class DockerRuntime(object):
    """Lightweight struct for our DockerImage configuration"""
    def __init__(self, id=id, image=None, command=None, host_config=None, default_params=None, server_registry_file=None):
        """
        Args:
            id: The short identifier for this runtime
            image: The full docker image name including a tag
            command: A list of commands to be passed to docker
            host_config: A dict that will be fed to docker.Client().create_host_config
            default_params: The default parameter values for this environment
            server_registry: A file containing a JSON dump of the server registry. The format will be runtime-specific.
        """
        self.id = id
        self.image = image
        self.command = command or []
        self.host_config = host_config or {}
        self.default_params = default_params or {}

        self._server_registry = None
        self._server_registry_file = server_registry_file

    @property
    def server_registry(self):
        if self._server_registry is None:
            with open(self._server_registry_file) as f:
                self._server_registry = json.load(f)
        return self._server_registry

    @property
    def _cli_flags(self):
        # Not everything maps in a straightforward way, e.g. cap_add => '--cap-add' but ipc_mode => '--ipc
        api_to_cli = {
            'ipc_mode': 'ipc'
        }

        cli_flags = []
        for api_key, api_value in self.host_config.items():
            if isinstance(api_value, (six.string_types, bool)):
                cli_values = [api_value]
            else:
                cli_values = api_value

            for cli_value in cli_values:
                if api_key in api_to_cli:
                    api_key = api_to_cli[api_key]
                cli_flag = '--{}'.format(api_key.replace('_', '-'))
                if isinstance(cli_value, bool):
                    # boolean flag, like --privileged
                    cli_flags += [cli_flag]
                else:
                    cli_flags += [cli_flag, cli_value]

        return cli_flags

    def cli_command(self, vnc_port, rewarder_port, extra_flags=[]):
        return ['docker', 'run',
           '-p', '{}:5900'.format(vnc_port),
           '-p', '{}:15900'.format(rewarder_port)] + \
           extra_flags + \
           self._cli_flags + \
           [self.image] + self.command


class WindowsRuntime(object):
    # TODO: Spawn windows runtimes (right now managed manually)
    def __init__(self, id=id, default_params=None):
        """
        Args:
            id: The short identifier for this runtime
        """
        self.id = id
        self.default_params = default_params


class Registry(object):
    def __init__(self):
        self.runtimes = collections.OrderedDict()

    def register_runtime(self, id, kind, **kwargs):
        if kind == "docker":
            self.runtimes[id] = DockerRuntime(id, **kwargs)
        elif kind == "windows":
            self.runtimes[id] = WindowsRuntime(id, **kwargs)
        else:
            raise error.Error("No runtime of kind {} . \n Valid options are ['docker']".format(kind))

    def runtime_spec(self, id):
        """
        id is a string describing the runtime, e.g 'flashgames

        Returns a configured DockerRuntime object
        """
        try:
            return self.runtimes[id]
        except KeyError:
            raise UnregisteredRuntime('No registered runtime with name: {}'.format(id))


registry = Registry()
register_runtime = registry.register_runtime
runtime_spec = registry.runtime_spec



import logging
import numpy as np
import os
from universe import pyprofile
import sys

from universe import error
from universe.vncdriver import server_messages

logger = logging.getLogger(__name__)

class PygletScreen(object):
    def __init__(self, bitmap=None):
        self._window = None
        self._is_updated = False
        self._height, self._width, _ = bitmap.shape
        self._initialize()
        self.update_rectangle(0, 0, self._width, self._height, bitmap)

    def flip(self):
        if not self._is_updated:
            return
        self._is_updated = False

        self._window.clear()
        self._window.switch_to()
        self._window.dispatch_events()
        self.texture.blit(0, 0)
        self._window.flip()

    def _initialize(self):
        if not os.environ.get('DISPLAY') and sys.platform.startswith('linux'):
            raise error.Error("Cannot render with mode='human' with no DISPLAY variable set.")

        import pyglet
        self._window = pyglet.window.Window(width=self._width, height=self._height, visible=True)
        self._window.dispatch_events()
        self.texture = pyglet.image.Texture.create(width=self._width, height=self._height)

    def update_rectangle(self, x, y, width, height, data):
        bytes = data.tobytes()
        pyprofile.incr('vncdriver.pyglet_screen.blit')
        pyprofile.incr('vncdriver.pyglet_screen.blit.bytes', len(bytes), unit=pyprofile.BYTES)
        import pyglet
        image = pyglet.image.ImageData(width, height, 'RGB', bytes, pitch=width * -3)
        self.texture.blit_into(image, x, self._height-height-y, 0)
        self._is_updated = True

    def apply(self, framebuffer_update):
        pyprofile.push('vncdriver.pyglet_screen.apply')
        for rect in framebuffer_update.rectangles:
            if isinstance(rect.encoding,
                          (server_messages.RAWEncoding, server_messages.ZRLEEncoding, server_messages.ZlibEncoding)):
                self.update_rectangle(rect.x, rect.y, rect.width, rect.height, rect.encoding.data)
            else:
                raise error.Error('Unrecognized encoding: {}'.format(rect.encoding))
        pyprofile.pop()



    # # TODO: we don't seem to be able to have multiple independent
    # # windows at once
    # def update_rectangle(self, x, y, width, height, data):
    #     self._update_rgbarray(x, y, width, height, update)


    # def copy_rectangle(self, src_x, src_y, x, y, width, height):
    #     assert self._window
    #     rectangle = self.texture.get_region(src_x, self._height-height-src_y, width, height)
    #     self.texture.blit_into(rectangle.get_image_data(), x, self._height-height-y, 0)

    # def fill_rectangle(self, x, y, width, height, color):
    #     import pyglet
    #     # While this technically works, it's super slow
    #     update = np.frombuffer(color, dtype=np.uint8)
    #     r, g, b = update[self._color_cycle]
    #     image_pattern = pyglet.image.SolidColorImagePattern(color=(r, g, b, 0))
    #     image = image_pattern.create_image(width, height)
    #     self.texture.blit_into(image, x, self._height-height-y, 0)

    # def commit(self):
    #     self._window.clear()
    #     self._window.switch_to()
    #     self.texture.blit(0, 0)

    #     self._is_updated = True



import json
import struct
import time

from gym.utils import atomic_write, closer

fbs_closer = closer.Closer()

class FBSWriter(object):
    def __init__(self, path):
        self._closed = False

        self.start = None
        self.stop = None

        self._id = fbs_closer.register(self)

        self.file = open(path, 'wb')
        # custom format: exactly the same as FBS 001.000 except:
        #
        # FBS 001.002
        # {line-of-json}
        # [length-byte, data, timestamp]...
        # \0\0\0\0 {line-of-json}
        self.file.write(b'FBS 001.002\n')

    def write(self, data):
        # Format:
        #
        # length
        # data
        # timestamp (4 bytes)

        if not data:
            return

        if self.start is not None:
            delta = int(1000 * (time.time() - self.start))
        else:
            delta = 0
            self.start = time.time()

            # Write metadata header
            self.file.write(json.dumps({'start': self.start}).encode('utf-8'))
            self.file.write(b'\n')

        length = struct.pack('!I', len(data))
        self.file.write(length)
        self.file.write(data)

        delta = struct.pack('!I', delta)
        self.file.write(delta)

    def _write_metadata(self):
        # Write metadata trailer
        null = struct.pack('!I', 0)
        self.file.write(null)
        self.file.write(json.dumps({'stop': self.stop}).encode('utf-8'))
        self.file.write(b'\n')

    def close(self):
        if self._closed:
            return
        self._closed = True

        fbs_closer.unregister(self._id)
        self.stop = time.time()
        self._write_metadata()
        self.file.close()

    def __del__(self):
        self.close()



class Remote(object):
    def __init__(self, handle, vnc_address, vnc_password, rewarder_address, rewarder_password, name=None):
        self.name = name
        self.handle = handle
        self.vnc_address = vnc_address
        self.vnc_password = vnc_password
        self.rewarder_address = rewarder_address
        self.rewarder_password = rewarder_password

    def __str__(self):
        return 'Remote<{}:{}>'.format(self.handle, self.name)

    def __repr__(self):
        return str(self)



import json
import logging
import pipes
import subprocess

class Error(Exception):
    pass

logger = logging.getLogger()

def pretty_command(command):
    return ' '.join(pipes.quote(c) for c in command)

def log_command(command, prefix=''):
    logger.info('%sExecuting: %s', prefix, pretty_command(command))

def check_call(command, *args, **kwargs):
    log_command(command)
    return subprocess.check_call(command, *args, **kwargs)

def popen(command, *args, **kwargs):
    log_command(command)
    return subprocess.Popen(command, *args, **kwargs)

def check_with_output(command, *args, **kwargs):
    log_command(command)
    proc = subprocess.Popen(command, *args, stdout=subprocess.PIPE, **kwargs)
    stdout, _ = proc.communicate()
    if proc.returncode != 0:
        raise Error('Command {} returned non-zero exit status {}'.format(command, proc.returncode))
    return stdout

def interpret_ready(pod):
    # status:
    # conditions:
    # - lastProbeTime: null
    #   lastTransitionTime: 2016-07-06T05:29:45Z
    #   message: 'containers with unready status: [xdummy xvnc vnc-atari]'
    #   reason: ContainersNotReady
    #   status: "False"
    #   type: Ready
    if 'conditions' not in pod['status']:
        return False

    ready = [c for c in pod['status']['conditions'] if c['type'] == 'Ready']
    if not ready:
        return False

    return ready[0]['status'] == 'True'

def interpret_ports(containers):
    # TODO: clean up hack
    try:
        recorder = containers['vnc-recorder']
    except KeyError:
        pass
    else:
        spec = recorder['ports'][0]
        assert spec['containerPort'] == 5899
        return spec['hostPort'], None

    app = [k for k in containers.keys() if k.startswith('vnc-')]
    assert len(app) == 1
    app = app[0]

    port_mapping = {}
    for spec in containers[app]['ports']:
        port_mapping[spec['containerPort']] = spec['hostPort']
    # vnc, rewarder
    return port_mapping[5900], port_mapping.get(15900)

class VNCEnvDiscovery(object):
    def __init__(self):
        self.context = 'sci'
        self.namespace = 'gym'
        self.kubectl = ['kubectl', '--context', self.context, '--namespace', self.namespace]

    def discover_batches(self):
        pods = check_with_output(self.kubectl + ['get', 'pods', '-o', 'json', '-l', 'type=universe'])
        pods = json.loads(pods)

        batches = {}
        for pod in pods['items']:
            if 'deletionTimestamp' in pod['metadata']:
                # Pod has been deleted!
                continue

            batch = pod['metadata']['labels']['batch']
            if batch not in batches:
                batches[batch] = {'count': 0}
            batches[batch]['count'] += 1
        return batches

    def discover(self, batch, force_ready=False):
        pods = check_with_output(self.kubectl + ['get', 'pods', '-o', 'json', '-l', 'type=universe', '-l', 'batch={}'.format(batch)])
        pods = json.loads(pods)

        if len(pods['items']) == 0:
            raise Error('Incorrect batch id: {}'.format(batch))

        remotes = []

        for pod in pods['items']:
            name = pod['metadata']['name']
            containers = {}
            for container in pod['spec']['containers']:
                containers[container['name']] = container
            vnc_port, rewarder_port = interpret_ports(containers)
            node = pod['spec'].get('nodeName')

            # Not scheduled on a node yet
            if node is None:
                if force_ready:
                    raise Error('Not all pods ready: {} is not scheduled on a node yet'.format(name))
                continue

            address = '{}:{}'.format(node, vnc_port)
            if rewarder_port is not None:
                address += '+{}'.format(rewarder_port)
            spec = {
                'name': name,
                'address': address,
                'ready': interpret_ready(pod),
            }
            remotes.append(spec)
        return remotes

vnc_env_discovery = VNCEnvDiscovery()
discover = vnc_env_discovery.discover
discover_batches = vnc_env_discovery.discover_batches



import gym
from gym import spaces
from universe import error

class Env(gym.Env):
    """Base class capable of handling vectorized environments.
    """
    metadata = {
        # This key indicates whether an env is vectorized (or, in the case of
        # Wrappers where autovectorize=True, whether they should automatically
        # be wrapped by a Vectorize wrapper.)
        'runtime.vectorized': True,
    }

    # Number of remotes. User should set this.
    n = None

class Wrapper(Env, gym.Wrapper):
    """Use this instead of gym.Wrapper iff you're wrapping a vectorized env,
    (or a vanilla env you wish to be vectorized).
    """
    # If True and this is instantiated with a non-vectorized environment, 
    # automatically wrap it with the Vectorize wrapper.
    autovectorize = True

    def __init__(self, env):
        super(Wrapper, self).__init__(env)
        if not env.metadata.get('runtime.vectorized'):
            if self.autovectorize:
                # Circular dependency :(
                from universe import wrappers
                env = wrappers.Vectorize(env)
            else:
                raise error.Error('This wrapper can only wrap vectorized envs (i.e. where env.metadata["runtime.vectorized"] = True), not {}. Set "self.autovectorize = True" to automatically add a Vectorize wrapper.'.format(env))

        self.env = env

    def _configure(self, **kwargs):
        super(Wrapper, self)._configure(**kwargs)
        assert self.env.n is not None, "Did not set self.env.n: self.n={} self.env={} self={}".format(self.env.n, self.env, self)
        self.n = self.env.n

class ObservationWrapper(Wrapper, gym.ObservationWrapper):
    pass

class RewardWrapper(Wrapper, gym.RewardWrapper):
    pass

class ActionWrapper(Wrapper, gym.ActionWrapper):
    pass




import logging
import os
from twisted.python.runtime import platform
from universe import vectorized

logger = logging.getLogger(__name__)

class Render(vectorized.Wrapper):
    def __init__(self, *args, **kwargs):
        if platform.isLinux() and not os.environ.get('DISPLAY'):
            self.renderable = False
        else:
            self.renderable = True
        self._observation = None
        super(Render, self).__init__(*args, **kwargs)

    def _configure(self, **kwargs):
        super(Render, self)._configure(**kwargs)
        self.metadata = self.metadata.copy()
        modes = self.metadata.setdefault('render.modes', [])
        if 'rgb_array' not in modes:
            modes.append('rgb_array')

    def _reset(self):
        observation_n = self.env.reset()
        self._observation = observation_n[0]
        return observation_n

    def _step(self, action_n):
        observation_n, reward_n, done_n, info_n = self.env.step(action_n)
        self._observation = observation_n[0]
        return observation_n, reward_n, done_n, info_n

    def _render(self, mode='human', *args, **kwargs):
        if not self.renderable and mode == 'human':
            return
        elif self.env is None:
            # Only when this breaks
            return
        elif mode == 'rgb_array':
            if self._observation is not None:
                observation = self._observation
                if isinstance(self._observation, dict):
                    observation = observation['vision']
                return observation
            else:
                return None
        # Could log, but no need.
        return self.env.render(mode=mode, *args, **kwargs)



from universe.vncdriver.screen.base import Screen
from universe.vncdriver.screen.numpy_screen import NumpyScreen
from universe.vncdriver.screen.pyglet_screen import PygletScreen
from universe.vncdriver.screen.screen_buffer import ScreenBuffer



import gym

class VNCObservationSpace(gym.Space):
    # For now, we leave the VNC ObservationSpace wide open, since
    # there isn't much use-case for this object.
    def contains(self, x):
        return True



import logging

from gym import monitoring
from universe.vectorized import core  # Cannot import vectorized directly without inducing a cycle

logger = logging.getLogger(__name__)

def Monitor(directory, video_callable=None, force=False, resume=False,
            write_upon_reset=False, uid=None, mode=None):
    class Monitor(core.Wrapper):
        def _configure(self, **kwargs):
            super(Monitor, self)._configure(**kwargs)

            # We have to wait until configure to set the monitor because we need the number of instances in a vectorized env
            self._start_monitor()

        def _start_monitor(self):
            # Circular dependencies :(
            from universe import wrappers
            # We need to maintain pointers to these to avoid them being
            # GC'd. They have a weak reference to us to avoid cycles.
            self._unvectorized_envs = [wrappers.WeakUnvectorize(self, i) for i in range(self.n)]

            # For now we only monitor the first env
            self._monitor = monitoring.MonitorManager(self._unvectorized_envs[0])
            self._monitor.start(directory, video_callable, force, resume,
                                write_upon_reset, uid, mode)

        def _step(self, action_n):
            self._monitor._before_step(action_n[0])
            observation_n, reward_n, done_n, info = self.env.step(action_n)
            done_n[0] = self._monitor._after_step(observation_n[0], reward_n[0], done_n[0], info)
            return observation_n, reward_n, done_n, info

        def _reset(self):
            self._monitor._before_reset()
            observation_n = self.env.reset()
            self._monitor._after_reset(observation_n[0])
            return observation_n

        def _close(self):
            super(Monitor, self)._close()
            self._monitor.close()

        def set_monitor_mode(self, mode):
            logger.info("Setting the monitor mode is deprecated and will be removed soon")
            self._monitor._set_mode(mode)

    return Monitor



import os
import re
import signal
import time

from universe import error
from universe.twisty import reactor
from twisted.internet import defer, protocol
import twisted.internet.error
import logging

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

class ConnectionTimer(protocol.Protocol):
    def connectionMade(self):
        self.transport.loseConnection()

def start(endpoint):
    start = time.time()
    return endpoint.connect(
        protocol.ClientFactory.forProtocol(ConnectionTimer)
    ).addCallback(lambda _: time.time() - start)

def measure_clock_skew(label, host):
    cmd = ['ntpdate', '-q', '-p', '8', host]
    extra_logger.info('[%s] Starting network calibration with %s', label, ' '.join(cmd))
    skew = Clockskew(label, cmd)
    # TODO: search PATH for this?
    process = reactor.spawnProcess(skew, '/usr/sbin/ntpdate', cmd, {})
    # process = reactor.spawnProcess(skew, '/bin/sleep', ['sleep', '2'], {})

    t = float(os.environ.get('UNIVERSE_NTPDATE_TIMEOUT', 20))
    def timeout():
        if process.pid:
            logger.error('[%s] %s call timed out after %ss; killing the subprocess. This is ok, but you could have more accurate timings by enabling UDP port 123 traffic to your env. (Alternatively, you can try increasing the timeout by setting environment variable UNIVERSE_NTPDATE_TIMEOUT=10.)', label, ' '.join(cmd), t)
            process.signalProcess(signal.SIGKILL)
            process.reapProcess()
    # TODO: make this part of the connection string
    reactor.callLater(t, timeout)
    return skew.deferred

class Clockskew(protocol.ProcessProtocol):
    def __init__(self, label, cmd):
        self.label = label
        self._cmd = cmd

        self.deferred = defer.Deferred()
        self.out = []
        self.err = []

    def outReceived(self, data):
        self.out.append(data)

    def errReceived(self, data):
        self.err.append(data)

    def processExited(self, reason):
        if isinstance(reason.value, twisted.internet.error.ProcessDone):
            out = b''.join(self.out).decode('utf-8')
            match = re.search('offset ([\d.-]+) sec', out)
            if match is not None:
                offset = float(match.group(1))
                self.deferred.callback(offset)
            else:
                self.deferred.errback(error.Error('Could not parse offset: %s', out))
        else:
            err = b''.join(self.err)
            self.deferred.errback(error.Error('{} failed with status {}: stderr={!r}'.format(self._cmd, reason.value.exitCode, err)))

class ConnectionTimerException(Exception):
    pass



import getpass
import logging
import os
import uuid

import universe
from gym.utils import reraise
from universe import error, pyprofile, rewarder, spaces, twisty, vectorized, vncdriver
from universe import remotes as remotes_module
from universe.envs import diagnostics
from universe.runtimes import registration
from universe.vncdriver import libvnc_session

# The Go driver is the most supported one. So long as the Go driver
# turns out to be easy to install, we'll continue forcing the Go
# driver here.
# noinspection PyUnresolvedReferences
import go_vncdriver

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

if os.environ.get('UNIVERSE_VNCDRIVER', None) == 'go':
    # Importing the Go driver early is desirable, so that people get
    # errors if it's not present. Also sometimes if go_vncdriver is
    # loaded after TensorFlow it will crash.
    import go_vncdriver

def default_client_id():
    return '{}-{}'.format(uuid.uuid4(), getpass.getuser())

def rewarder_session(which):
    if which is None:
        which = rewarder.RewarderSession

    if isinstance(which, type):
        return which
    else:
        raise error.Error('Invalid RewarderSession driver: {!r}'.format(which))

def go_vncdriver():
    import go_vncdriver
    return go_vncdriver.VNCSession

def py_vncdriver():
    return vncdriver.VNCSession

def libvnc_vncdriver():
    return libvnc_session.LibVNCSession

def vnc_session(which=None):
    # Short circuit so long as we're forcing the Go driver. Other code
    # left behind for the future if we need to support the other
    # drivers again.
    if isinstance(which, type):
        # Used in the tests to pass a custom VNC driver
        return which

    logger.info('Using the golang VNC implementation')
    return go_vncdriver()

    if which is None:
        which = os.environ.get('UNIVERSE_VNCDRIVER')

    if isinstance(which, type):
        return which
    if which == 'go':
        logger.info('Using the golang VNC implementation')
        return go_vncdriver()
    elif which == 'py':
        logger.info('Using the Python VNC implementation')
        return py_vncdriver()
    elif which == 'libvnc':
        logger.info('Using the libvnc VNC implementation')
        return libvnc_vncdriver()
    elif which is None:
        try:
            go = go_vncdriver()
            logger.debug('Using golang VNC implementation')
            return go
        except ImportError as e:
            logger.info("Go driver failed to import: {}".format(e))
            logger.info("Using pure Python vncdriver implementation. Run 'pip install go-vncdriver' to install the more performant Go implementation. Optionally set the environment variable UNIVERSE_VNCDRIVER='go' to force its use.")
            return py_vncdriver()
    else:
        raise error.Error('Invalid VNCSession driver: {!r}'.format(which))

def build_observation_n(visual_observation_n, info_n):
    observation_n = []
    for visual, info in zip(visual_observation_n, info_n):
        text = info.pop('env.text', [])
        obs = {
            'vision': visual,
            'text': text,
        }
        if 'env.generic' in info:
            obs['generic'] = info.pop('env.generic')
        observation_n.append(obs)
    return observation_n

class VNCEnv(vectorized.Env):
    metadata = {
        'render.modes': ['human'], # we wrap with a Render which can render to rgb_array
        'semantics.async': True,
        'semantics.autoreset': True,
        'video.frames_per_second' : 60,
        'runtime.vectorized': True,
    }

    def __init__(self, fps=None, probe_key=None):
        self.metadata = dict(self.metadata)
        if fps is not None:
            self.metadata['video.frames_per_second'] = fps

        self._started = False

        self.observation_space = spaces.VNCObservationSpace()
        self.action_space = spaces.VNCActionSpace()

        self._seed_value = None
        self._remotes_manager = None

        self._probe_key = probe_key or 0xbeef1

        self.vnc_session = None
        self.rewarder_session = None

        self._send_actions_over_websockets = False
        self._skip_network_calibration = False


    def _seed(self, seed):
        self._seed_value = seed
        return [seed]

    def _configure(self, remotes=None,
                   client_id=None,
                   start_timeout=None, docker_image=None,
                   ignore_clock_skew=False, disable_action_probes=False,
                   vnc_driver=None, vnc_kwargs={},
                   rewarder_driver=None,
                   replace_on_crash=False, allocate_sync=True,
                   observer=False, api_key=None,
                   record=False,
    ):
        """Standard Gym hook to configure the environment.

        Args:

          ignore_clock_skew (bool): Assume remotes are on the same machine as us,
            for the purposes of diagnostics measurement.

            If true, we skip measuring the clock skew over the network,
            and skip generating diagnostics which rely on it.

            True when used by the rewarder to measure latency between
            the VNC frame and its calculation of reward for that
            frame.  In this case we share a common clock with the env
            generating the VNC frame, so we don't need to send/receive
            probes.  Clock skew is zero in this case.

            False when remotes are potentially different machines
            (such as an agent, or a demonstrator), and we will be
            sending probe keys and measuring network ping rountrip
            times to calculate clock skew.
        """
        if self._started:
            raise error.Error('{} has already been started; cannot change configuration now.'.format(self))

        universe.configure_logging()

        twisty.start_once()

        if self.spec is not None:
            runtime = registration.runtime_spec(self.spec.tags['runtime'])
            # Let the user manually set the docker_image version
            if docker_image:
                # TODO: don't support this option?
                runtime.image = docker_image
        else:
            runtime = None

        if remotes is None:
            remotes = os.environ.get('GYM_VNC_REMOTES', '1')

        if client_id is None:
            client_id = default_client_id()

        self.remote_manager, self.n = remotes_module.build(
            client_id=client_id,
            remotes=remotes, runtime=runtime, start_timeout=start_timeout,
            api_key=api_key,
            use_recorder_ports=record,
        )
        self.connection_names = [None] * self.n
        self.connection_labels = [None] * self.n
        self.crashed = {}

        self.allow_reconnect = replace_on_crash and self.remote_manager.supports_reconnect
        if self.remote_manager.connect_vnc:
            cls = vnc_session(vnc_driver)
            vnc_kwargs.setdefault('start_timeout', self.remote_manager.start_timeout)
            if runtime == 'gym-core':
                vnc_kwargs.setdefault('encoding', 'zrle')
            else:
                vnc_kwargs.setdefault('encoding', 'tight')
                vnc_kwargs.setdefault('fine_quality_level', 50)
                vnc_kwargs.setdefault('subsample_level', 2)
            # Filter out None values, since some drivers may not handle them correctly
            vnc_kwargs = {k: v for k, v in vnc_kwargs.items() if v is not None}
            logger.info('Using VNCSession arguments: %s. (Customize by running "env.configure(vnc_kwargs={...})"', vnc_kwargs)
            self.vnc_kwargs = vnc_kwargs
            self.vnc_session = cls()
        else:
            self.vnc_session = None

        self._observer = observer
        if self.remote_manager.connect_rewarder:
            cls = rewarder_session(rewarder_driver)
            self.rewarder_session = cls()
        else:
            self.rewarder_session = None

        if ignore_clock_skew:
            logger.info('Printed stats will ignore clock skew. (This usually makes sense only when the environment and agent are on the same machine.)')

        if self.rewarder_session or ignore_clock_skew:
            # Don't need rewarder session if we're ignoring clock skew
            if self.spec is not None:
                metadata_encoding = self.spec.tags.get('metadata_encoding')
            else:
                metadata_encoding = None
            self.diagnostics = diagnostics.Diagnostics(self.n, self._probe_key, ignore_clock_skew, metadata_encoding=metadata_encoding, disable_action_probes=disable_action_probes)
        else:
            self.diagnostics = None

        self._reset_mask()
        self._started = True

        self.remote_manager.allocate([str(i) for i in range(self.n)], initial=True)
        if allocate_sync:
            # Block until we've fulfilled n environments
            self._handle_connect(n=self.n)
        else:
            # Handle any backends which synchronously fufill their
            # allocation.
            self._handle_connect()

    def connect(self, i, name, vnc_address, rewarder_address, vnc_password=None, rewarder_password=None):
        logger.info('[%s] Connecting to environment: vnc://%s password=%s. Use the browser-based VNC client: http://%s/viewer/?password=%s', name, vnc_address, vnc_password, rewarder_address, vnc_password)

        extra_logger.info('[%s] Connecting to environment details: vnc_address=%s vnc_password=%s rewarder_address=%s rewarder_password=%s', name, vnc_address, vnc_password, rewarder_address, rewarder_password)
        self.connection_names[i] = name
        self.connection_labels[i] = '{}:{}'.format(name, vnc_address)
        if self.vnc_session is not None:
            kwargs = {
                'name': name,
                'address': vnc_address,
                'password': vnc_password,
            }
            kwargs.update(self.vnc_kwargs)

            try:
                self.vnc_session.connect(**kwargs)
            except TypeError as e:
                reraise(suffix="(HINT: this error was while passing arguments to the VNCSession driver: {})".format(kwargs))

            # TODO: name becomes index:pod_id
            # TODO: never log index, just log name

        if self.rewarder_session is not None:
            if self.spec is not None:
                env_id = self.spec.id
            else:
                env_id = None

            if self._seed_value is not None:
                # Once we use a seed, we clear it so we never
                # accidentally reuse the seed. Seeds are an advanced
                # feature and aren't supported by most envs in any
                # case.
                seed = self._seed_value[i]
                self._seed_value[i] = None
            else:
                seed = None

            assert rewarder_password, "Missing rewarder password: {}".format(rewarder_password)
            network = self.rewarder_session.connect(
                name=name, address=rewarder_address,
                seed=seed, env_id=env_id,
                fps=self.metadata['video.frames_per_second'],
                password=rewarder_password,
                label=self.connection_labels[i],
                start_timeout=self.remote_manager.start_timeout,
                observer=self._observer,
                skip_network_calibration=self._skip_network_calibration
            )
        else:
            network = None

        if self.diagnostics is not None:
            self.diagnostics.connect(i, network, label=name)
        self.crashed.pop(i, None)

    def _close(self, i=None):
        if i is not None:
            name = self.connection_names[i]
            if self.rewarder_session:
                self.rewarder_session.close(name)
            if self.vnc_session:
                self.vnc_session.close(name)
            if self.diagnostics:
                self.diagnostics.close(i)
            self.mask.close(i)
            self.connection_names[i] = None
            self.connection_labels[i] = None
        else:
            if hasattr(self, 'rewarder_session') and self.rewarder_session:
                self.rewarder_session.close()
            if hasattr(self, 'vnc_session') and self.vnc_session:
                self.vnc_session.close()
            if hasattr(self, 'diagnostics') and self.diagnostics:
                self.diagnostics.close()
            if hasattr(self, 'remotes_manager') and self._remotes_manager:
                self._remotes_manager.close()

    def _reset(self):
        self._handle_connect()

        if self.rewarder_session:
            self.rewarder_session.reset()
        self._reset_mask()
        return [None] * self.n

    def _reset_mask(self):
        self.mask = Mask(self.connection_labels, initially_masked=bool(self.rewarder_session))

    def _pop_rewarder_session(self, peek_d):
        with pyprofile.push('vnc_env.VNCEnv.rewarder_session.pop'):
            reward_d, done_d, info_d, err_d = self.rewarder_session.pop(peek_d=peek_d)

        reward_n = []
        done_n = []
        info_n = []
        err_n = []
        for name in self.connection_names:
            reward_n.append(reward_d.get(name, 0))
            done_n.append(done_d.get(name, False))
            info_n.append(info_d.get(name, {'env_status.disconnected': True}))
            err_n.append(err_d.get(name))
        return reward_n, done_n, info_n, err_n

    def _step_vnc_session(self, compiled_d):
        if self._send_actions_over_websockets:
            self.rewarder_session.send_action(compiled_d, self.spec.id)
            vnc_action_d = {}
        else:
            vnc_action_d = compiled_d

        with pyprofile.push('vnc_env.VNCEnv.vnc_session.step'):
            observation_d, info_d, err_d = self.vnc_session.step(vnc_action_d)

        observation_n = []
        info_n = []
        err_n = []
        for name in self.connection_names:
            observation_n.append(observation_d.get(name))
            info_n.append(info_d.get(name))
            err_n.append(err_d.get(name))

        return observation_n, info_n, err_n

    def _compile_actions(self, action_n):
        compiled_n = []
        peek_d = {}
        try:
            for i, action in enumerate(action_n):
                compiled = []
                compiled_n.append(compiled)
                for event in action:
                    # Handle any special control actions
                    if event == spaces.PeekReward:
                        name = self.connection_names[i]
                        peek_d[name] = True
                        continue

                    # Do a generic compile
                    compiled.append(compile_action(event))
        except Exception as e:
            raise error.Error('Could not compile actions. Original error: {} ({}). action_n={}'.format(e, type(e), action_n))
        else:
            return compiled_n, peek_d

    def _action_d(self, action_n):
        action_d = {}
        for i, action in enumerate(action_n):
            action_d[self.connection_names[i]] = action
        return action_d

    def _step(self, action_n):
        self._handle_connect()

        # Compile actions to something more palatable by drivers
        # written in other language.
        action_n, peek_d = self._compile_actions(action_n)

        # We pop from the rewarder session first since we need to
        # determine if any of the current VNC actions need to be
        # masked. (If the environment is resetting, we should
        # definitely not send it any actions.)
        #
        # It's a bit counterintuitive to check for rewards first,
        # since we haven't submitted the actions yet, but keep in mind
        # that everything here is asynchronous!
        if self.rewarder_session:
            reward_n, done_n, info_n, err_n = self._pop_rewarder_session(peek_d)
        else:
            reward_n = done_n = [None] * self.n
            info_n = [{} for _ in range(self.n)]
            err_n = [None] * self.n

        action_mask = observation_mask = self.mask.maintain_mask(done_n, info_n)

        # Drop any actions to resetting environments.
        #
        # We pass the mask to add_probe so it doesn't try to schedule
        # probes which are potentially harmful and will just time out
        # anyway.
        self.mask.apply_to_actions(action_n, info_n, action_mask)
        # Send our actions and return the current framebuffer
        if self.vnc_session:
            if self.diagnostics:
                self.diagnostics.clear_probes_when_done(done_n)
                self.diagnostics.add_probe(action_n, action_mask)
            action_d = self._action_d(action_n)

            visual_observation_n, obs_info_n, vnc_err_n = self._step_vnc_session(action_d)
            # Merge in any keys from the observation
            self._propagate_obs_info(info_n, obs_info_n)
        else:
            visual_observation_n = [None] * self.n
            vnc_err_n = [None] * self.n

        observation_n = build_observation_n(visual_observation_n, info_n)
        self.mask.apply_to_return(observation_n, reward_n, done_n, info_n, observation_mask)

        self._handle_initial_n(observation_n, reward_n)
        self._handle_err_n(err_n, vnc_err_n, info_n, observation_n, reward_n, done_n)
        self._handle_crashed_n(info_n)

        return observation_n, reward_n, done_n, {'n': info_n}

    def _handle_initial_n(self, observation_n, reward_n):
        if self.rewarder_session is None:
            return

        for i, reward in enumerate(reward_n):
            if reward is None:
                # Index hasn't come up yet, so ensure the observation
                # is blanked out.
                observation_n[i] = None

    def _handle_err_n(self, err_n, vnc_err_n, info_n, observation_n=None, reward_n=None, done_n=None):
        # Propogate any errors upwards.
        for i, (err, vnc_err) in enumerate(zip(err_n, vnc_err_n)):
            if err is None and vnc_err is None:
                # All's well at this index.
                continue

            if observation_n is not None:
                observation_n[i] = None
                done_n[i] = True

            # Propagate the error
            if err is not None and vnc_err is not None:
                # Both the rewarder and vnc failed at the same
                # time. What are the odds?
                info_n[i]['error'] = 'Both VNC and rewarder sessions failed: {}; {}'.format(vnc_err, err)
            elif err is not None:
                info_n[i]['error'] = 'Rewarder session failed: {}'.format(err)
            else:
                info_n[i]['error'] = 'VNC session failed: {}'.format(vnc_err)

            if self.allow_reconnect:
                logger.info('[%s] Making a call to the allocator to replace crashed index: %s', self.connection_names[i], info_n[i]['error'])
                self.remote_manager.allocate([str(i)])

            self.crashed[i] = self.connection_names[i]
            self._close(i)

    def _handle_connect(self, n=None):
        # Connect to any environments which are ready
        for remote in self.remote_manager.pop(n=n):
            if remote.name is not None:
                name = '{}:{}'.format(remote.handle, remote.name)
            else:
                name = remote.handle

            self.connect(
                int(remote.handle), name=name,
                vnc_address=remote.vnc_address, vnc_password=remote.vnc_password,
                rewarder_address=remote.rewarder_address, rewarder_password=remote.rewarder_password)

    def _handle_crashed_n(self, info_n):
        # for i in self.crashed:
        #     info_n[i]['env_status.crashed'] = True

        if self.allow_reconnect:
            return

        if len(self.crashed) > 0:
            errors = {}
            for i, info in enumerate(info_n):
                if 'error' in info:
                    errors[self.crashed[i]] = info['error']

            if len(errors) == 0:
                raise error.Error('{}/{} environments have crashed. No error key in info_n: {}'.format(len(self.crashed), self.n, info_n))
            else:
                raise error.Error('{}/{} environments have crashed! Most recent error: {}'.format(len(self.crashed), self.n, errors))

    def _propagate_obs_info(self, info_n, obs_info_n):
        for obs_info, info in zip(obs_info_n, info_n):
            # obs_info keys take precedence over info keys
            if obs_info is not None:
                info.update(obs_info)

    def _render(self, mode='human', close=False):
        if close:
            # render(close) is not currently supported by the Go VNCSession
            return

        if mode is 'human' and self.vnc_session is not None:
            if self.connection_names[0]:
                self.vnc_session.render(self.connection_names[0])

    def __str__(self):
        return 'VNCEnv<{}>'.format(self.spec.id)

class Mask(object):
    """Blocks the agent from interacting with the environment while the
    environment is resetting.

    On the boundaries:

    - Mask will *drop* actions to environments which have just started
      resetting (i.e. returning done=True in this iteration and have
      env_state=resetting).

    - Mask will *allow* actions to environments which have just
      finished resetting (i.e. their env_state=running).

    - Mask will *allow* rewards from environments which have just
      started resetting (i.e. returning done=True in this iteration
      and have env_state=resetting), but mask observations from them.

    - Mask will *allow* observations from environments which have just
      finished resetting (i.e. their env_state is running)
    """
    def __init__(self, connection_labels, initially_masked=True):
        self.connection_labels = connection_labels
        self.n = len(connection_labels)

        self.episode_ids = [None] * self.n
        if initially_masked:
            self.mask = [None] * self.n
        else:
            self.mask = [True] * self.n

    def close(self, i):
        self.mask[i] = None
        self.episode_ids[i] = None

    def maintain_mask(self, done_n, info_n):
        for i, ok in enumerate(self.mask):
            if info_n[i].get('peek'):
                env_state = info_n[i].get('env_status.peek.env_state', 'resetting')
                episode_id = info_n[i].get('env_status.peek.episode_id')

                if info_n[i].get('env_status.episode_id') != episode_id:
                    completed_episode_id = info_n[i].get('env_status.episode_id')
                else:
                    completed_episode_id = None
            else:
                env_state = info_n[i].get('env_status.env_state', 'resetting')
                episode_id = info_n[i].get('env_status.episode_id')
                completed_episode_id = info_n[i].get('env_status.complete.episode_id')

            # Either:
            # 1. The env is currently masked (not ok)
            # 2. We have an active episode (self.episode_ids[i])
            # 3. We didn't connect the rewarder (done_n[i] is None)
            assert not ok or self.episode_ids[i] is not None or done_n[i] is None, "ok={} episode_id={} i={}".format(ok, episode_id, i)
            if not ok and env_state == 'running':
                extra_logger.info('[%s] Episode began: episode_id=%s env_state=%s', self.connection_labels[i], episode_id, env_state)
                self.mask[i] = True
                # this should change only for initial reset
                self.episode_ids[i] = episode_id
            elif ok and self.episode_ids[i] != episode_id and env_state == 'running':
                extra_logger.info('[%s] Episode ended (and began, so not masking): episode_id=%s->%s env_state=%s', self.connection_labels[i], completed_episode_id, episode_id, env_state)
                self.episode_ids[i] = episode_id
            elif ok and self.episode_ids[i] != episode_id:
                extra_logger.info('[%s] Episode ended: episode_id=%s->%s env_state=%s', self.connection_labels[i], completed_episode_id, episode_id, env_state)
                self.mask[i] = False
                self.episode_ids[i] = episode_id
        return self.mask

    def apply_to_actions(self, action_n, info_n, mask):
        for i, ok in enumerate(mask):
            if ok:
                continue

            action_n[i] = []
            info_n[i]['mask.masked.action'] = True
        return self.mask

    def apply_to_return(self, observation_n, reward_n, done_n, info_n, observation_mask):
        # Next, mask any environments which are resetting. We are
        # guaranteed to get done=True messages prior to getting the
        # v0.env.describe message telling us it's resetting, so the
        # conservative route (block upon done=True, unblock upon
        # v0.env.describe with env_state=running) locks us out of the
        # maximum surface area of environment reset possible.
        for i, ok in enumerate(observation_mask):
            if ok:
                continue

            if len(observation_n[i]['text']) > 0 and ok is False:
                logger.warn('[%s] WARNING: Masking text observation for environment %d: %r. This means we received text data before the environment finished resetting; the text observation has been lost. This is not expected and should be reported.', self.connection_labels[i], i, observation_n[i]['text'])

            observation_n[i] = None
            info_n[i]['mask.masked.observation'] = True

def compile_action(event):
    if isinstance(event, tuple):
        if event[0] == 'KeyEvent':
            name, down = event[1:]
            return spaces.KeyEvent.by_name(name, down=down).compile()
        elif event[0] == 'PointerEvent':
            x, y, buttonmask = event[1:]
            return spaces.PointerEvent(x, y, buttonmask).compile()
    else:
        return event.compile()



import logging
import time
from universe import pyprofile, vectorized

logger = logging.getLogger(__name__)

class Timer(vectorized.Wrapper):
    """
Calcultae how much time was spent actually doing work.  Display result
via pyprofile.
"""

    def _reset(self):
        with pyprofile.push('vnc_env.Timer.reset'):
            return self.env.reset()

    def _step(self, action_n):
        start = time.time()
        with pyprofile.push('vnc_env.Timer.step'):
            observation_n, reward_n, done_n, info = self.env.step(action_n)

        # Calculate how much time was spent actually doing work
        sleep = info.get('stats.throttle.sleep')
        if sleep is None or sleep < 0:
            sleep = 0
        pyprofile.timing('vnc_env.Timer.step.excluding_sleep', time.time() - start - sleep)
        return observation_n, reward_n, done_n, info



import logging
from gym import configuration

universe_logger = logging.getLogger('universe')
universe_logger.setLevel(logging.INFO)

extra_logger = logging.getLogger('universe.extra')
extra_logger.setLevel(logging.INFO)

if hasattr(configuration, '_extra_loggers'):
    configuration._extra_loggers.append(universe_logger)
    configuration._extra_loggers.append(extra_logger)



import threading
from twisted.python.runtime import platform

# On OSX, we should use kqueue rather than the default select
# backend. (Proximal issue is that select only can handle a limited
# number of file descriptors.)
#
# Based off twisted.internet.default
def _get_reactor(platform):
    try:
        if platform.isLinux():
            try:
                from twisted.internet import epollreactor
                cls = epollreactor.EPollReactor
            except ImportError:
                from twisted.internet import pollreactor
                cls = pollreactor.PollReactor
        elif platform.isMacOSX():
            from twisted.internet import kqreactor
            cls = kqreactor.KQueueReactor
        elif platform.getType() == 'posix' and not platform.isMacOSX():
            from twisted.internet import pollreactor
            cls = pollreactor.PollReactor
        else:
            from twisted.internet import selectreactor
            cls = selectreactor.SelectReactor
    except ImportError:
        from twisted.internet import selectreactor
        cls = selectreactor.SelectReactor
    return cls()

class TwistedThread(threading.Thread):
    started = False
    daemon = True

    @classmethod
    def start_once(cls):
        if cls.started:
            return
        cls.started = True

        instance = cls(name='Twisted')
        instance.start()

    def run(self):
        reactor.run(installSignalHandlers=False)

reactor = _get_reactor(platform)
start_once = TwistedThread.start_once



from gym.spaces import prng

class Hardcoded(object):
    def __init__(self, actions):
        self.actions = actions

    def contains(self, action):
        return action in self.actions

    def sample(self):
        i = prng.np_random.randint(len(self.actions))
        return self.actions[i]

    def __getitem__(self, i):
        return self.actions[i]



import logging

from universe.vncdriver.vnc_session import VNCSession
from universe.vncdriver.vnc_client import client_factory
from universe.vncdriver.screen import NumpyScreen, PygletScreen

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



from universe.spaces.hardcoded import Hardcoded
from universe.spaces.vnc_action_space import VNCActionSpace
from universe.spaces.vnc_event import VNCEvent, KeyEvent, PointerEvent
from universe.spaces.vnc_observation_space import VNCObservationSpace

from universe.spaces.diagnostics import PeekReward



import collections
import fastzbarlight
import itertools
import logging
from multiprocessing import pool
import numpy as np
import time
import threading
# import psutil
import sys
from collections import namedtuple
from gym.utils import reraise

import re
from universe import error, pyprofile, spaces

# TODO: prefix the loggers

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

def show(ob):
    from PIL import Image
    Image.fromarray(ob).show()

def standard_error(ary, axis, scale=1):
    ary = np.array(ary) * scale
    if len(ary) > 1:
        return np.std(ary, axis=axis) / np.sqrt(len(ary) - 1)
    else:
        return np.std(ary, axis=axis)

def extract_timestamp(observation):
    total = 0
    for byte in observation[0]:
        total = 256 * total + byte
    for byte in observation[1]:
        total = 256 * total + byte

    timestamp = total/1000.
    return timestamp

class MetadataDecoder(object):
    @classmethod
    def build(cls, metadata_encoding, pool, qr_pool, label):
        metadata_encoding = metadata_encoding.copy()
        type = metadata_encoding.pop('type')
        if type == 'qrcode':
            return QRCodeMetadataDecoder(label=label, pool=pool, qr_pool=qr_pool, **metadata_encoding)
        elif type == 'pixels':
            return PixelsMetadataDecoder(label=label)
        else:
            raise error.Error('Invalid encoding: {}'.format(type))

class AsyncDecode(object):
    pool = None

    def __init__(self, pool, qr_pool, method, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

        self._last_img = None

        self.method = method
        self.results = []
        self.deque = collections.deque()
        self.pool = pool
        self.qr_pool = qr_pool

    def __call__(self, img, available_at):
        # Choose the return value
        if len(self.deque) > 0 and self.deque[0].ready():
            last = self.deque.popleft()
            res = last.get()
            if res is not None:
                pyprofile.timing('vnc_env.diagnostics.async_decode.latency', time.time() - res['available_at'])
        else:
            res = False

        pyprofile.gauge('vnc_env.diagnostics.async_decode.queue_depth', len(self.deque))

        # Just grayscale it by keeping only one component. Should be
        # good enough as this region is black and white anyway.
        grayscale = img[self.y:self.y+self.height, self.x:self.x+self.width, 0]

        # Apply processing if needed
        match = np.array_equal(self._last_img, grayscale)
        if not match:
            pyprofile.incr('vnc_env.diagnostics.async_decode.schedule')
            # sneakily copy if numpy hasn't, so it can be cached
            self._last_img = np.ascontiguousarray(grayscale)
            async = self.qr_pool.apply_async(self.method, (self._last_img, time.time(), available_at))
            self.deque.append(async)
        else:
            pyprofile.incr('vnc_env.diagnostics.async_decode.cache_hit')

        return res

class QRCodeMetadataDecoder(MetadataDecoder):
    def __init__(self, pool, qr_pool, x, y, width, height, label):
        self.flag_synchronous = False

        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.label = label

        self.decode = AsyncDecode(pool, qr_pool, self._decode, x, y, width, height)

    def _decode(self, observation, start, available_at):
        # This method gets wrapped by AsyncDecode.__call__
        with pyprofile.push('vnc_env.diagnostics.QRCodeMetadataDecoder.qr_code_scanner'):

            encoded = fastzbarlight.qr_code_scanner(observation.tobytes(), self.width, self.height)
        if encoded is None:
            # Failed to parse!
            return
        if encoded.startswith(b'v1:'):
            encoded = encoded.decode('utf-8')
            if len(encoded) != len('v1:') + 12 + 12:
                raise error.Error('Bad length for metadata from enviroment: {}'.format(encoded))

            encoded = encoded[len('v1:'):]
            last_update = int(encoded[:12], 16) / 1000.0
            last_action = int(encoded[12:24], 16) / 1000.

            return {
                # Timestamp on the image
                'now': last_update,
                # When the last probe was received
                'probe_received_at': last_action,
                'processing_start': start,
                'processing_end': time.time(),
                'available_at': available_at,
            }
        else:
            raise error.Error('Bad version string for metadata from environment: {}'.format(encoded))


class PixelsMetadataDecoder(MetadataDecoder):
    def __init__(self, label):
        self.flag_synchronous = True

        self.anchor = np.array([
            [(0x12, 0x34, 0x56), (0x78, 0x90, 0xab)],
            [(0x23, 0x45, 0x67), (0x89, 0x0a, 0xbc)],
        ], dtype=np.uint8)

        self.location = None
        self.last_search_metadata = 0
        self.label = label

    def _check_location(self, observation, location):
        y, x = location
        return np.all(observation[y:y+2, x:x+2] == self.anchor)

    def _find_metadata_location(self, observation):
        ys, xs = np.where(np.all(observation == self.anchor[0, 0], axis=-1))
        if len(ys) == 0:
            extra_logger.info('[%s] Could not find metadata anchor pixel', self.label)
            return False

        # TODO: handle multiple hits
        assert len(ys) == 1
        location = (ys[0], xs[0])

        assert self._check_location(observation, location)
        extra_logger.info('[%s] Found metadata anchor pixel: %s', self.label, location)
        return location

    def _should_search_metadata(self):
       return time.time() - self.last_search_metadata > 1

    def decode(self, observation, available_at=None):
        start = time.time()

        # metadata pixel location hasn't been initialized or it has moved
        if not self.location or not self._check_location(observation,
                                                         self.location):
            # only search for metadata occasionally
            if self._should_search_metadata():
                self.location = self._find_metadata_location(observation)
                self.last_search_metadata = time.time()

        if not self.location:
            return False  # False translates to None in DiagnosticsInstance

        y, x = self.location
        now = extract_timestamp(observation[y, x+2:x+4])
        probe_received_at = extract_timestamp(observation[y, x+4:x+6])

        return {
            'now': now,
            'probe_received_at': probe_received_at,
            'processing_start': start,
            'processing_end': time.time(),
            'available_at': available_at,
        }

class Diagnostics(object):
    def __init__(self, n, probe_key, ignore_clock_skew=False, metadata_encoding=None, disable_action_probes=False):
        # Each QR code takes about 1ms (and updates at 5fps). We do
        # our best to ensure the QR is processed in time for the next
        # step call (n/16 would put us right at the threshold).
        self.pool = pool.ThreadPool(max(int(n/4), 1))
        self.qr_pool = pool.ThreadPool(max(int(n/8), 1))
        self.lock = threading.RLock()

        self.instance_n = [None] * n
        self.ignore_clock_skew = ignore_clock_skew
        self.disable_action_probes = disable_action_probes

        self.metadata_encoding = metadata_encoding

        self.update(probe_key=probe_key, metadata_encoding=metadata_encoding)

    # only used in flashgames right now
    def update(self, probe_key, metadata_encoding):
        self.probe_key = probe_key
        self.metadata_encoding = metadata_encoding

        for instance in self.instance_n:
            if instance is not None:
                instance.update(probe_key=self.probe_key, metadata_encoding=self.metadata_encoding)

    def connect(self, i, network=None, label=None):
        # This should technically be synchronized
        self.instance_n[i] = DiagnosticsInstance(i, network, self.probe_key, self.ignore_clock_skew, self.metadata_encoding, disable_action_probes=self.disable_action_probes, qr_pool=self.qr_pool, pool=self.pool, label=label)

    def close(self, i=None):
        if i is not None:
            self.instance_n[i] = None
        else:
            self.pool.close()
            self.qr_pool.close()
            for i in range(len(self.instance_n)):
                self.close(i)
            self.instance_n = None

    def add_probe(self, action_n, mask_n):
        if self.disable_action_probes:
            return

        for instance, action, mask in zip(self.instance_n, action_n, mask_n):
            # Important that masking prevents us from adding probes. (This
            # avoids us e.g. filling in backticks into text boxes as the
            # environment boots.)
            if mask and instance:
                instance.add_probe(action)

    def add_metadata(self, observation_n, info_n, available_at=None):
        """Mutates the info_n dictionary."""
        if self.instance_n is None:
            return

        with pyprofile.push('vnc_env.diagnostics.Diagnostics.add_metadata'):
            async = self.pool.imap_unordered(
                self._add_metadata_i,
                zip(self.instance_n, observation_n, info_n, [available_at] * len(observation_n)))
            list(async)

    def _add_metadata_i(self, args):
        instance, observation, info, now = args
        if instance is None or observation is None:
            return
        instance.add_metadata(observation, info, now)

    def extract_metadata(self, observation_n):
        return [instance._extract_metadata(observation)
                for instance, observation in zip(self.instance_n, observation_n)]

    def clear_probes_when_done(self, done_n):
        for instance, done in zip(self.instance_n, done_n):
            if done:
                instance.clear_probe()

class DiagnosticsInstance(object):
    anchor = np.array([
        [(0x12, 0x12, 0x12), (0x78, 0x78, 0x78)],
        [(0x23, 0x23, 0x23), (0x89, 0x89, 0x89)],
    ], dtype=np.uint8)
    zero_clock_skew = np.zeros([2])

    def __init__(self, i, network, probe_key, ignore_clock_skew, metadata_encoding, disable_action_probes, pool, qr_pool, label=None):
        '''
        network - either Network() object used to get clock skew, or None.

                  If None, we skip measuring clock skew, and skip measuring
                  diagnostics which rely on clock skew.
        '''
        if network is None:
            assert ignore_clock_skew
        self.ignore_clock_skew = ignore_clock_skew

        self.label = label
        self.i = i
        self.network = network

        self.probe_sent_at = None # local time
        self.probe_received_at = None # remote time
        self.action_latency_skewed = None
        self.last_observation_timestamp = None
        self.disable_action_probes = disable_action_probes

        self.pool = pool
        self.qr_pool = qr_pool
        self.could_read_metadata = None
        self.update(probe_key=probe_key, metadata_encoding=metadata_encoding)

    def update(self, probe_key, metadata_encoding):
        self.probe = [
                spaces.KeyEvent(probe_key, down=True).compile(),
                spaces.KeyEvent(probe_key, down=False).compile(),
        ]

        if metadata_encoding is not None:
            self.metadata_decoder = MetadataDecoder.build(metadata_encoding, pool=self.pool, qr_pool=self.qr_pool, label=self.label)
        else:
            self.metadata_decoder = None

    def clear_probe(self):
        self.probe_sent_at = None
        self.probe_received_at = None

    def add_probe(self, action):
        if self.network is not None and not self.network.active():
            return

        if self.probe_sent_at is not None and self.probe_sent_at + 10 < time.time():
            extra_logger.warn('[%s] Probe to determine action latency timed out (was sent %s). (This is harmless, but worth knowing about.)', self.label, self.probe_sent_at)
            self.probe_sent_at = None
        if self.probe_sent_at is None:
            extra_logger.debug('[%s] Sending out new action probe: %s', self.label, self.probe)
            self.probe_sent_at = time.time()
            action += self.probe
        assert self.probe_sent_at is not None

    def add_metadata(self, observation, info, available_at=None):
        """Extract metadata from a pixel observation and add it to the info
        """
        observation = observation['vision']
        if observation is None: return
        if self.network is not None and not self.network.active():
            return
        elif self.metadata_decoder is None:
            return
        elif observation is None:
            return
        # should return a dict with now/probe_received_at keys
        with pyprofile.push('vnc_env.diagnostics.DiagnosticsInstance.add_metadata.decode'):
            metadata = self.metadata_decoder.decode(observation, available_at=available_at)

        if metadata is False:
            # No metadata ready, though it doesn't mean parsing failed
            metadata = None
        elif metadata is None:
            if self.could_read_metadata:
                self.could_read_metadata = False
                extra_logger.info('[%s] Stopped being able to read metadata (expected when environment resets)', self.label)
        elif not self.could_read_metadata:
            self.could_read_metadata = True
            extra_logger.info('[%s] Started being able to read metadata', self.label)

        if self.metadata_decoder.flag_synchronous and metadata is not None:
            info['diagnostics.image_remote_time'] = metadata['now']

        local_now = time.time()

        if self.network is None:
            # Assume the clock skew is zero. Should only be run on the
            # same machine as the VNC server, such as the universe
            # instance inside of the environmenth containers.
            real_clock_skew = self.zero_clock_skew
        else:
            # Note: this is a 2-length vector of (min, max), so anything added to
            # it is also going to be a 2-length vector.
            # Most of the diagnostics below are, but you have to look carefully.
            real_clock_skew = self.network.reversed_clock_skew()

        # Store real clock skew here
        info['stats.gauges.diagnostics.clock_skew'] = real_clock_skew
        if self.ignore_clock_skew:
            clock_skew = self.zero_clock_skew
        else:
            clock_skew = real_clock_skew

        if metadata is not None:
            # We'll generally update the observation timestamp infrequently
            if self.last_observation_timestamp == metadata['now']:
                delta = None
            else:
                # We just got a new timestamp in the observation!
                self.last_observation_timestamp = metadata['now']
                observation_now = metadata['now']
                delta = observation_now - metadata['available_at']

                # Subtract *local* time it was received from the *remote* time
                # displayed. Negate and reverse order to fix time ordering.
                info['stats.gauges.diagnostics.lag.observation'] = -(delta + clock_skew)[[1, 0]]

            # if self.network is None:
            #     # The rest of diagnostics need the network, so we're done here
            #     return

            probe_received_at = metadata['probe_received_at']
            if probe_received_at == 0 or self.disable_action_probes:
                # Happens when the env first starts
                self.probe_received_at = None
            elif self.probe_received_at is None: # this also would work for the equality case
                self.probe_received_at = probe_received_at
            elif self.probe_received_at != probe_received_at and self.probe_sent_at is None:
                logger.info('[%s] Probe is marked as received at %s, but probe_sent_at is None. This is surprising. (HINT: do you have multiple universe instances talking to the same environment?)', self.label, probe_received_at)
            elif self.probe_received_at != probe_received_at:
                extra_logger.debug('[%s] Next probe received: old=%s new=%s', self.label, self.probe_received_at, probe_received_at)
                self.probe_received_at = probe_received_at
                # Subtract the *local* time we sent it from the *remote* time it was received
                self.action_latency_skewed = probe_received_at - self.probe_sent_at
                self.probe_sent_at = None

            if self.action_latency_skewed:
                action_lag = self.action_latency_skewed + clock_skew
                self.action_latency_skewed = None
            else:
                action_lag = None
            info['stats.gauges.diagnostics.lag.action'] = action_lag

        local_now = time.time()
        # Look at when the remote believed it parsed the score (not
        # all envs send this currently).
        #
        # Also, if we received no new rewards, then this values is
        # None. This could indicate a high reward latency (bad,
        # uncommon), or that the agent is calling step faster than new
        # rewards are coming in (good, common).
        remote_score_now = info.get('rewarder.lag.observation.timestamp')
        if remote_score_now is not None:
            delta = remote_score_now - local_now
            info['stats.gauges.diagnostics.lag.reward'] = -(delta + clock_skew)[[1, 0]]

        # Look at when the remote send the message, so we know how
        # long it's taking for messages to get to us.
        rewarder_message_now = info.get('reward_buffer.remote_time')
        if rewarder_message_now:
            delta = rewarder_message_now - local_now
            info['stats.gauges.diagnostics.lag.rewarder_message'] = -(delta + clock_skew)[[1, 0]]


def extract_n_m(dict_n_m, key):
    output = []
    for dict_n in dict_n_m:
        layer = []
        for dict in dict_n:
            layer.append(dict[key])
        output.append(layer)
    return np.array(output)


# class ChromeProcessInfo(object):
#     proc_regex = re.compile('.*(chrome|Chrome|nacl_helper).*')

#     def add_system_stats(self, info, now):
#         """TODO: This needs be moved to universe-envs and run there. Otherwise it only works if the env and agent
#             are on the same machine. In addition a new rpc call, rpc.env.diagnostics, should be added to return
#             data to the agent periodically.
#         """
#         start = time.time()

#         # CPU
#         cpu_percent = psutil.cpu_percent()
#         info['diagnostics.env.cpu.percent'] = cpu_percent
#         cpu_cores_percent = psutil.cpu_percent(percpu=True)
#         num_cores = len(cpu_cores_percent)
#         info['diagnostics.env.cpu.percent.all_cores'] = cpu_percent / num_cores
#         info['diagnostics.env.cpu.percent.each_core'] = cpu_cores_percent
#         info['diagnostics.env.cpu.num_cores'] = num_cores

#         # MEMORY
#         mem = psutil.virtual_memory()
#         info['diagnostics.env.memory.percent'] = mem.percent
#         info['diagnostics.env.memory.total'] = mem.total
#         info['diagnostics.env.memory.available'] = mem.available

#         # NETWORK
#         if self.last_measured_at is not None:
#             elapsed_ms = (now - self.last_measured_at) * 1000.
#             current = psutil.net_io_counters()
#             dl = (current.bytes_recv - self.system_network_counters.bytes_recv) / elapsed_ms
#             ul = (current.bytes_sent - self.system_network_counters.bytes_sent) / elapsed_ms
#             info['diagnostics.env.network.download_bytes_ps'] = dl * 1000.
#             info['diagnostics.env.network.upload_bytes_ps'] = ul * 1000.
#             self.system_network_counters = current

#         # CHROME
#         if self.chrome_last_measured_at is None or (time.time() - self.chrome_last_measured_at) > 30:
#             # Fetch every 30 seconds
#             self.chrome_last_measured_at = time.time()
#             logger.info("Measuring Chrome process statistics")
#             chrome_info = ChromeProcessInfo()
#             chrome_info = best_effort(chrome_info.fetch, num_cores)
#             if chrome_info is not None:
#                 self.chrome_info = chrome_info

#         if self.chrome_info is not None:
#             self._populate_chrome_info(self.chrome_info, info)

#         # TODO: Add GPU stats

#         pyprofile.push('diagnostics.system_stats')

#     def _populate_chrome_info(self, chrome_info, info):
#         pyprofile.push('diagnostics.chrome_process_info.process_iter')
#         pyprofile.push('diagnostics.chrome_process_info.total')
#         info['diagnostics.chrome.age'] = chrome_info.age
#         info['diagnostics.chrome.cpu.time'] = chrome_info.cpu_time
#         info['diagnostics.chrome.cpu.percent'] = chrome_info.cpu_percent
#         info['diagnostics.chrome.cpu.percent.all_cores'] = chrome_info.cpu_percent_all_cores
#         info['diagnostics.chrome.cpu.percent.all_cores_all_time'] = chrome_info.cpu_percent_all_cores_all_time
#         info['diagnostics.chrome.num_processes'] = len(chrome_info.processes)

#     def __init__(self):
#         self.cpu_time = 0.
#         self.cpu_percent = 0.
#         self.min_create_time = None
#         self.visited_pids = set()
#         self.processes = []
#         self.time_to_get_procs = None
#         self.total_time_to_measure = None
#         self.age = None
#         self.cpu_percent_all_cores_all_time = None
#         self.cpu_percent_all_cores = None

#     def fetch(self, num_cores):
#         start = time.time()
#         start_process_iter = time.time()
#         procs = list(psutil.process_iter())
#         self.time_to_get_procs = time.time() - start_process_iter
#         for proc in procs:
#             try:
#                 name = proc.name()
#                 if self.proc_regex.match(name):
#                     self._fetch_single(proc, name)
#                     # N.B. Don't read children. defunct processes make this take 4ever.
#                     # Child processes are all uncovered by initial scan.
#             except (psutil.AccessDenied, psutil.NoSuchProcess) as e:
#                 pass
#         self.total_time_to_measure = time.time() - start
#         if self.min_create_time is None:
#             self.age = 0
#         else:
#             self.age = time.time() - self.min_create_time
#         self.cpu_percent_all_cores_all_time = 100. * self.cpu_time / (self.age * num_cores)
#         self.cpu_percent_all_cores = self.cpu_percent / num_cores
#         return self

#     def _fetch_single(self, proc, name):
#         if proc.pid in self.visited_pids:
#             return
#         try:
#             cpu_times = proc.cpu_times()
#             cpu_percent = proc.cpu_percent()
#             created = proc.create_time()
#             if self.min_create_time is None:
#                 self.min_create_time = created
#             else:
#                 self.min_create_time = min(created, self.min_create_time)

#             cpu_time = cpu_times.user + cpu_times.system
#             proc_info = namedtuple('proc_info', 'name cpu_time cpu_percent created age')
#             proc_info.name = name
#             proc_info.cpu_time = cpu_time
#             proc_info.cpu_percent = cpu_percent
#             proc_info.created = created
#             proc_info.age = time.time() - created
#             proc_info.pid = proc.pid
#             self.processes.append(proc_info)

#             # Totals
#             self.cpu_time += cpu_time
#             self.cpu_percent += cpu_percent
#             self.visited_pids.add(proc.pid)


#         except (psutil.AccessDenied, psutil.NoSuchProcess) as e:
#             pass



from __future__ import absolute_import

import base64
import logging
import os
import pipes
import sys
import threading
import uuid

import docker
import six.moves.urllib.parse as urlparse
from gym.utils import closer
from universe import error
from universe.remotes import healthcheck, remote
from universe import error, utils
from universe.remotes.compose import container, log_printer, progress_stream

logger = logging.getLogger(__name__)

docker_closer = closer.Closer()

def random_alphanumeric(length=14):
    buf = []
    while len(buf) < length:
        entropy = base64.encodestring(uuid.uuid4().bytes).decode('utf-8')
        bytes = [c for c in entropy if c.isalnum()]
        buf += bytes
    return ''.join(buf)[:length]

def pretty_command(command):
    return ' '.join(pipes.quote(c) for c in command)

class DockerManager(object):
    def __init__(self, runtime, n, reuse=False, start_timeout=None):
        super(DockerManager, self).__init__()

        self.runtime = runtime

        self.supports_reconnect = False
        self.connect_vnc = True
        self.connect_rewarder = True
        self._assigner = PortAssigner(reuse=reuse)

        self._popped = False

        self.lock = threading.Lock()
        self.envs = []

        self._n = n
        if start_timeout is None:
            start_timeout = 2 * self._n + 5
        self.start_timeout = start_timeout
        self._start()

    def allocate(self, handles, initial=False, params={}):
        self._handles = handles

    def pop(self, n=None):
        """Call from main thread. Returns the list of newly-available (handle, env) pairs."""
        if self._popped:
            assert n is None
            return []
        self._popped = True

        envs = []
        for i, instance in enumerate(self.instances):
            env = remote.Remote(
                handle=self._handles[i],
                vnc_address='{}:{}'.format(instance.host, instance.vnc_port),
                vnc_password='openai',
                rewarder_address='{}:{}'.format(instance.host, instance.rewarder_port),
                rewarder_password='openai',
            )
            envs.append(env)
        return envs

    def _start(self):
        self.instances = [DockerInstance(self._assigner, self.runtime, label=str(i)) for i in range(self._n)]

        [instance.start() for instance in self.instances]
        self.start_logging(self.instances)
        self.healthcheck(self.instances)

    def close(self):
        with self.lock:
            [instance.close() for instance in self.instances]

    def start_logging(self, instances):
        containers = [instance._container for instance in instances]
        labels = [str(instance.label) for instance in instances]
        if all(instance.reusing for instance in instances):
            # All containers are being reused, so only bother showing
            # a subset of the backlog.
            tail = 0
        else:
            # At least one container is new, so just show
            # everything. It'd be nice to have finer-grained control,
            # but this would require patching the log printer.
            tail = 'all'
        log_printer.build(containers, labels, log_args={'tail': tail})

    def healthcheck(self, instances):
        # Wait for boot
        healthcheck.run(
            ['{}:{}'.format(instance.assigner.info['host'], instance.vnc_port) for instance in instances],
            ['{}:{}'.format(instance.assigner.info['host'], instance.rewarder_port) for instance in instances],
            start_timeout=30,
        )

def get_client():
    info = {}
    host = os.environ.get('DOCKER_HOST')

    client_api_version = os.environ.get('DOCKER_API_VERSION')
    if not client_api_version:
        client_api_version = "auto"

    # IP to use for started containers
    if host:
        info['host'] = urlparse.urlparse(host).netloc.split(':')[0]
    else:
        info['host'] = 'localhost'

    verify = os.environ.get('DOCKER_TLS_VERIFY') == '1'
    if verify: # use TLS
        assert_hostname = None
        cert_path = os.environ.get('DOCKER_CERT_PATH')
        if cert_path:
            client_cert = (os.path.join(cert_path, 'cert.pem'), os.path.join(cert_path, 'key.pem'))
            ca_cert = os.path.join(cert_path, 'ca.pem')
        else:
            client_cert = ca_cert = None

        tls_config = docker.tls.TLSConfig(
            client_cert=client_cert,
            ca_cert=ca_cert,
            verify=verify,
            assert_hostname=assert_hostname,
        )
        return docker.Client(base_url=host, tls=tls_config, version=client_api_version), info
    else:
        return docker.Client(base_url=host, version=client_api_version), info

class PortAssigner(object):
    def __init__(self, reuse=False):
        self.reuse = reuse
        self.instance_id = 'universe-' + random_alphanumeric(length=6)
        self.client, self.info = get_client()
        self._refresh_ports()

    def _refresh_ports(self):
        ports = {}
        for container in self.client.containers():
            for port in container['Ports']:
                # {u'IP': u'0.0.0.0', u'Type': u'tcp', u'PublicPort': 5000, u'PrivatePort': 500}
                if port['Type'] == 'tcp' and 'PublicPort' in port:
                    ports[port['PublicPort']] = container['Id']
        self._ports = ports
        self._next_port = 5900

    def allocate_ports(self, num):
        if self.reuse and self._next_port in self._ports:
            vnc_id = self._ports[self._next_port]
            rewarder_id = self._ports.get(self._next_port+10000)

            # Reuse an existing docker container if it exists
            if (self._next_port+10000) not in self._ports:
                raise error.Error("Port {} was allocated but {} was not. This indicates unexpected state with spun-up VNC docker instances.".format(self._next_port, self._next_port+1))
            elif vnc_id != rewarder_id:
                raise error.Error("Port {} is exposed from {} while {} is exposed from {}. Both should come from a single Docker instance running your environment.".format(vnc_id, self._next_port, rewarder_id, self._next_port+10000))

            base = self._next_port
            self._next_port += 1
            return base, base+10000, vnc_id
        elif not self.reuse:
            # Otherwise, allocate find the lowest free pair of
            # ports. This doesn't work for the reuse case since on
            # restart we won't remember where we spun up our
            # containers.
            while self._next_port in self._ports or (self._next_port+10000) in self._ports:
                self._next_port += 1

        base = self._next_port
        self._next_port += 1

        # And get started!
        return base, base+10000, None

class DockerInstance(object):
    def __init__(self, assigner, runtime, label='main'):
        self._docker_closer_id = docker_closer.register(self)

        self.label = label
        self.assigner = assigner
        self.name='{}-{}'.format(self.assigner.instance_id, self.label),

        self.runtime = runtime

        self._container_id = None
        self._closed = False

        self._container = None

        self.host = self.assigner.info['host']
        self.vnc_port = None
        self.rewarder_port = None
        self.reusing = None
        self.started = False

    def start(self, attempts=None):
        if attempts is None:
            # If we're reusing, we don't scan through ports for a free
            # one.
            if not self.assigner.reuse:
                attempts = 5
            else:
                attempts = 1

        for attempt in range(attempts):
            self._spawn()
            e = self._start()
            if e is None:
                return
        raise error.Error('[{}] Could not start container after {} attempts. Last error: {}'.format(self.label, attempts, e))

    def _spawn(self):
        if self.runtime.image is None:
            raise error.Error('No image specified')
        assert self._container_id is None

        self.vnc_port, self.rewarder_port, self._container_id = self.assigner.allocate_ports(2)
        if self._container_id is not None:
            logger.info('[%s] Reusing container %s on ports %s and %s', self.label, self._container_id[:12], self.vnc_port, self.rewarder_port)
            self.reusing = True
            self.started = True
            return

        self.reusing = False
        logger.info('[%s] Creating container: image=%s. Run the same thing by hand as: %s',
                    self.label,
                    self.runtime.image,
                    pretty_command(self.runtime.cli_command(self.vnc_port, self.rewarder_port)))
        try:
            container = self._spawn_container()
        except docker.errors.NotFound as e:
            # Looks like we need to pull the image
            assert 'No such image' in e.explanation.decode('utf-8'), 'Expected NotFound error message message to include "No such image", but it was: {}. This is probably just a bug in this assertion and the assumption was incorrect'.format(e.explanation)

            logger.info('Image %s not present locally; pulling', self.runtime.image)
            self._pull_image()
            # Try spawning again
            container = self._spawn_container()

        self._container_id = container['Id']

    def _pull_image(self):
        output = self.client.pull(self.runtime.image, stream=True)
        return progress_stream.get_digest_from_pull(
            progress_stream.stream_output(output, sys.stdout))

        # docker-compose uses this:
        # try:
        # except StreamOutputError as e:
        #     if not ignore_pull_failures:
        #         raise
        #     else:
        #         log.error(six.text_type(e))

    def _spawn_container(self):
        # launch instance, and refresh if error
        container = self.client.create_container(
            image=self.runtime.image,
            command=self.runtime.command,
            # environment=self.runtime.environment,
            name=self.name,
            host_config=self.client.create_host_config(
                port_bindings={
                    5900: self.vnc_port,
                    15900: self.rewarder_port,
                 },
                **self.runtime.host_config),
            labels={
                'com.openai.automanaged': 'true',
            }
        )
        return container

    def _start(self):
        # Need to start up the container!
        if not self.started:
            logger.debug('[%s] Starting container: id=%s', self.label, self._container_id)
            try:
                self.client.start(container=self._container_id)
            except docker.errors.APIError as e:
                if 'port is already allocated' in str(e.explanation):
                    logger.info('[%s] Could not start container: %s', self.label, e)
                    self._remove()
                    return e
                else:
                    raise
            else:
                self.started = True

        self._container = container.Container.from_id(self.client, self._container_id)
        return None

    def _remove(self):
        logger.info("Killing and removing container: id=%s. (If this command errors, you can always kill all automanaged environments on this Docker daemon via: docker rm -f $(docker ps -q -a -f 'label=com.openai.automanaged=true')", self._container_id)
        self.client.remove_container(container=self._container_id, force=True)
        self._container_id = None

    def __del__(self):
        self.close()

    def close(self):
        if self._closed:
            return

        docker_closer.unregister(self._docker_closer_id)

        # Make sure 1. we were the onse who started it, 2. it's
        # actually been started, and 3. we're meant to kill it.
        if self._container_id and not self.assigner.reuse:
            self._remove()

        self._closed = True

    @property
    def client(self):
        return self.assigner.client

if __name__ == '__main__':
    logging.getLogger().setLevel(logging.INFO)
    from universe.runtimes import registration

    # docker run --name test --rm -ti -p 5900:5900 -p 15900:15900 quay.io/openai/universe.gym-core
    instance = DockerManager(
        runtime=registration.runtime_spec('gym-core'),
        n=2,
    )
    instance.start()
    import ipdb;ipdb.set_trace()



import logging
import os

from twisted.internet import defer, endpoints

from universe import error, utils
from universe.twisty import reactor
from universe.vncdriver import screen, vnc_client

PYGAME_INSTALLED = None
def load_pygame():
    global PYGAME_INSTALLED, pygame
    if PYGAME_INSTALLED is not None:
        return

    try:
        import pygame
        PYGAME_INSTALLED = True
    except ImportError:
        PYGAME_INSTALLED = False

logger = logging.getLogger(__name__)


class LibVNCSession(object):
    def __init__(self, remotes, error_buffer, encoding=None, compress_level=None, fine_quality_level=None, subsample_level=None):
        """compress_level: 0-9 [9 is highest compression]
        fine_quality_level: 0-100 [100 is best quality]
        subsample_level: 0-3 [0 is best quality]

        Lots of references for this, but
        https://github.com/TurboVNC/turbovnc/blob/master/doc/performance.txt
        is decent.
        """

        load_pygame()
        import libvncdriver

        if encoding is None:
            encoding = os.environ.get('LIBVNC_ENCODING', 'tight')
        if compress_level is None:
            compress_level = int(os.environ.get('LIBVNC_COMPRESS_LEVEL', '0'))
        if fine_quality_level is None:
            fine_quality_level = int(os.environ.get('LIBVNC_FINE_QUALITY_LEVEL', '100'))
        if subsample_level is None:
            subsample_level = int(os.environ.get('LIBVNC_SUBSAMPLE_LEVEL', '0'))

        if not hasattr(libvncdriver, 'VNCSession'):
            raise error.Error('''
 *=================================================*
|| libvncdriver is not installed                   ||
|| Try installing with "pip install libvncdriver"  ||
|| or use the go or python driver by setting       ||
|| UNIVERSE_VNCDRIVER=go                                ||
|| UNIVERSE_VNCDRIVER=py                                ||
 *=================================================*''')
        logger.info("Using libvncdriver's %s encoding" % encoding)
        self.driver = libvncdriver.VNCSession(
            remotes=remotes,
            error_buffer=error_buffer,
            encoding=encoding,
            compress_level=compress_level,
            fine_quality_level=fine_quality_level,
            subsample_level=subsample_level,
        )
        self.screen = None
        self.render_called_once = False
        if PYGAME_INSTALLED:
            pygame.init()

    def flip(self):
        return self._guard(self.driver.flip)

    def step(self, action):
        return self.driver.step(action)

    def render(self):
        self._guard(self._render)

    def _guard(self, fn):
        try:
            return fn()
        except (KeyboardInterrupt, SystemExit):
            self.close()

    def _render(self):
        self.before_render()
        if not PYGAME_INSTALLED:
            return
        # For some reason pygame wants X and Y swapped
        aray, n = self.driver.flip()
        if self.screen is None:
            self.screen = pygame.display.set_mode(aray[0].shape[:2][::-1])
        surf = pygame.surfarray.make_surface(aray[0].swapaxes(0, 1))
        rect = surf.get_rect()
        self.screen.blit(surf, rect)
        pygame.display.flip()

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.close()

    def before_render(self):
        if not self.render_called_once:
            self.render_called_once = True
            if not PYGAME_INSTALLED:
                logger.warn('''
 *================================================================*
||                                                                ||
|| Rendering disabled when using libvnc without pygame installed. ||
|| Consider viewing over VNC or running "pip install pygame".     ||
||                                                                ||
 *================================================================*''')


    def close(self):
        if PYGAME_INSTALLED:
            pygame.quit()
        self.driver.close()



import os
import yaml

from universe.runtimes.registration import register_runtime

with open(os.path.join(os.path.dirname(__file__), '../runtimes.yml')) as f:
    spec = yaml.load(f)

# If you have a local repo, do something like
# export OPENAI_DOCKER_REPO=docker.openai.com  (this one only for openai folks)
docker_repo = os.environ.get('OPENAI_DOCKER_REPO', 'quay.io/openai')

register_runtime(
    id='gym-core',
    kind='docker',
    image=docker_repo + '/universe.gym-core:{}'.format(spec['gym-core']['tag']),
)

register_runtime(
    id='flashgames',
    kind='docker',
    image=docker_repo + '/universe.flashgames:{}'.format(spec['flashgames']['tag']),
    host_config={
        'privileged': True,
        'cap_add': ['SYS_ADMIN'],
        'ipc_mode': 'host',
    },
    default_params={'cpu': 3.9, 'livestream_url': None},
    server_registry_file=os.path.join(os.path.dirname(__file__), 'flashgames.json'),
)

register_runtime(
    id='world-of-bits',
    kind='docker',
    image=docker_repo + '/universe.world-of-bits:{}'.format(spec['world-of-bits']['tag']),
    host_config={
        'privileged': True,
        'cap_add': ['SYS_ADMIN'],
        'ipc_mode': 'host'
    })

register_runtime(
    id='vnc-windows',
    kind='windows',
)

del spec



import gym
from universe import envs, error, spaces, vectorized

def atari_vnc(up=False, down=False, left=False, right=False, z=False):
    return [spaces.KeyEvent.by_name('up', down=up),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right),
            spaces.KeyEvent.by_name('down', down=down),
            spaces.KeyEvent.by_name('z', down=z)]

def slither_vnc(space=False, left=False, right=False):
    return [spaces.KeyEvent.by_name('space', down=space),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right)]

def racing_vnc(up=False, left=False, right=False):
    return [spaces.KeyEvent.by_name('up', down=up),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right)]

def platform_vnc(up=False, left=False, right=False, space=False):
    return [spaces.KeyEvent.by_name('up', down=up),
            spaces.KeyEvent.by_name('left', down=left),
            spaces.KeyEvent.by_name('right', down=right),
            spaces.KeyEvent.by_name('space', down=space)]

class SafeActionSpace(vectorized.Wrapper):
    """
Recall that every universe environment receives a list of VNC events as action.
There exist many environments for which the set of relevant action is much smaller
and is known.   For example, Atari environments have a modest number of keys,
so this wrapper, when applied to an Atari environment will reduce its action space.
Doing so is very convenient for research, since today's RL algorithms rely on random
exploration, which is hurt by small action spaces.  As our algorithms get better
and we switch to using the raw VNC commands, this wrapper will become less important.
"""

    def __init__(self, env):
        super(SafeActionSpace, self).__init__(env)

        if self.spec.tags.get('runtime') == 'gym-core':
            self.action_space = gym_core_action_space(self.spec._kwargs['gym_core_id'])
        elif self.spec is None:
            pass
        elif self.spec.id == 'internet.SlitherIO-v0' or self.spec.id == 'internet.SlitherIOErmiyaEskandaryBot-v0' or self.spec.id == 'internet.SlitherIOEasy-v0':
            self.action_space = spaces.Hardcoded([
                slither_vnc(left=True),
                slither_vnc(right=True),
                slither_vnc(space=True),
                slither_vnc(left=True, space=True),
                slither_vnc(right=True, space=True),
            ])
        elif self.spec.id in ['flashgames.DuskDrive-v0']:
            # TODO: be more systematic
            self.action_space = spaces.Hardcoded([
                racing_vnc(up=True),
                racing_vnc(left=True),
                racing_vnc(right=True),
            ])
        elif self.spec.id in ['flashgames.RedBeard-v0']:
            self.action_space = spaces.Hardcoded([
                platform_vnc(up=True),
                platform_vnc(left=True),
                platform_vnc(right=True),
                platform_vnc(space=True),
            ])

def gym_core_action_space(gym_core_id):
    spec = gym.spec(gym_core_id)

    if spec.id == 'CartPole-v0':
        return spaces.Hardcoded([
            [spaces.KeyEvent.by_name('left', down=True)],
            [spaces.KeyEvent.by_name('left', down=False)],
        ])
    elif spec._entry_point.startswith('gym.envs.atari:'):
        actions = []
        env = spec.make()
        for action in env.unwrapped.get_action_meanings():
            z = 'FIRE' in action
            left = 'LEFT' in action
            right = 'RIGHT' in action
            up = 'UP' in action
            down = 'DOWN' in action
            translated = atari_vnc(up=up, down=down, left=left, right=right, z=z)
            actions.append(translated)
        return spaces.Hardcoded(actions)
    else:
        raise error.Error('Unsupported env type: {}'.format(spec.id))



import json
import logging
import os
import re
import requests
import six
import six.moves.urllib.parse as urlparse
import threading
import time

import gym
from gym import scoreboard
from gym.utils import reraise

from universe import error, utils
from universe.remotes import remote

if six.PY2:
    import Queue as queue
else:
    import queue

logger = logging.getLogger(__name__)
extra_logger = logging.getLogger('universe.extra.'+__name__)

# Using gym for this
_api_key = 'tyytjgq3envte2j9yv2e-{}'.format(os.environ.get('OPENAI_USER', os.environ.get('USER')))
allocator_base = 'http://allocator.sci.openai-tech.com'
# gym_base_url = 'http://api.gym.sci.openai-tech.com'

class Stop(Exception):
    pass

class RequestError(Exception):
    def __init__(self, message, status_code=None):
        super(RequestError, self).__init__(message)
        self.message = message
        if status_code is not None:
            self.status_code = status_code

class FatalError(RequestError):
    pass

class AllocatorManager(threading.Thread):
    daemon = True

    def __init__(self, client_id, base_url=allocator_base,
                 address_type=None, start_timeout=None, api_key=None,
                 runtime_id=None, tag=None, params=None, placement=None,
                 use_recorder_ports=False,
    ):
        super(AllocatorManager, self).__init__()
        self.label = 'AllocatorManager'

        self.supports_reconnect = True
        self.connect_vnc = True
        self.connect_rewarder = True

        if address_type is None: address_type = 'public'
        if address_type not in ['public', 'pod', 'private']:
            raise error.Error('Bad address type specified: {}. Must be public, pod, or private.'.format(address_type))

        self.tag = tag
        self.client_id = client_id
        self.address_type = address_type

        if start_timeout is None:
            start_timeout = 20 * 60
        self.start_timeout = start_timeout
        self.params = params
        self.placement = placement
        self.use_recorder_ports = use_recorder_ports

#         if base_url is None:
#             base_url = scoreboard.api_base
#         if base_url is None:
#             base_url = gym_base_url
#         if api_key is None:
#             api_key = scoreboard.api_key
#         if api_key is None:
#             raise gym.error.AuthenticationError("""You must provide an OpenAI Gym API key.

# (HINT: Set your API key using "gym.scoreboard.api_key = .." or "export OPENAI_GYM_API_KEY=..."). You can find your API key in the OpenAI Gym web interface: https://gym.openai.com/settings/profile.""")

        if api_key is None:
            api_key = _api_key
        self._requestor = AllocatorClient(self.label, api_key, base_url=base_url)
        self.base_url = base_url

        # These could be overridden on a per-allocation basis, if you
        # want heterogeoneous envs. We don't support those currently
        # in the higher layers, but this layer could support it
        # easily.
        self.runtime_id = runtime_id
        self.tag = tag

        self.pending = {}

        self.error_buffer = utils.ErrorBuffer()
        self.requests = queue.Queue()
        self.ready = queue.Queue()

        self._reconnect_history = {}
        self._sleep = 1

    @classmethod
    def from_remotes(cls, client_id, remotes, runtime_id, start_timeout, tag, api_key, use_recorder_ports):
        parsed = urlparse.urlparse(remotes)
        if not (parsed.scheme == 'http' or parsed.scheme == 'https'):
            raise error.Error('AllocatorManager must start with http:// or https://: {}'.format(remotes))

        base_url = parsed.scheme + '://' + parsed.netloc
        if parsed.path:
            base_url += '/' + parsed.path
        query = urlparse.parse_qs(parsed.query)

        n = query.get('n', [1])[0]
        cpu = query.get('cpu', [None])[0]
        if cpu is not None:
            cpu = float(cpu)
        placement = query.get('address', ['public'])[0]

        params = {}
        if tag is not None: params['tag'] = tag
        if cpu is not None: params['cpu'] = cpu

        return cls(client_id=client_id, runtime_id=runtime_id, base_url=base_url, start_timeout=start_timeout, params=params, placement=placement, api_key=api_key, use_recorder_ports=use_recorder_ports), int(n)

    def pop(self, n=None):
        """Call from main thread. Returns the list of newly-available (handle, env) pairs."""
        self.error_buffer.check()

        envs = []

        if n is None:
            while True:
                try:
                    envs += self.ready.get(block=False)
                except queue.Empty:
                    break
        else:
            sync_timeout = 10 * 60
            start = time.time()

            wait_time = 1
            while len(envs) < n:
                try:
                    extra_logger.info('[%s] Waiting for %d envs, currently at %d, sleeping for %d', self.label, n, len(envs), wait_time)
                    envs += self.ready.get(timeout=wait_time)
                except queue.Empty:
                    self.error_buffer.check()
                wait_time = min(wait_time * 2, 30)
                delta = time.time() - start
                if delta > sync_timeout:
                    raise FatalError("Waited %.0fs to obtain envs, timeout was %.0fs. (Obtained %d/%d envs.)" % (delta, sync_timeout, len(envs), n))

        return envs

    def allocate(self, handles, initial=False, params={}):
        """Call from main thread. Initiate a request for more environments"""
        assert all(re.search('^\d+$', h) for h in handles), "All handles must be numbers: {}".format(handles)
        self.requests.put(('allocate', (handles, initial, params)))

    def close(self):
        self.requests.put(('close', ()))

    def run(self):
        try:
            self._run()
        except Stop:
            pass
        except Exception as e:
            self.error_buffer.record(e)

    def _run(self):
        while True:
            self._process_requests()
            self._poll()

    def _process_requests(self):
        while True:
            try:
                method, args = self.requests.get(timeout=self._sleep)
            except queue.Empty:
                break
            else:
                if method == 'allocate':
                    handles, initial, params = args
                    self._allocate(handles, initial, params)
                elif method == 'close':
                    raise Stop

    def _allocate(self, handles, initial, params):
        self._sleep = 1

        _params = self.params.copy()
        _params.update(params)

        for handle in handles:
            history = self._reconnect_history.get(handle, [])
            history.append(time.time())
            floor = time.time() - 5 * 60
            history = [entry for entry in history if entry > floor]
            if len(history) > 5:
                raise error.Error('Tried reallocating a fresh remote at index {} a total of {} times in the past 5 minutes (at {}). Please examine the logs to determine why the remotes keep failing.'.format(handle, len(history), history))
            self._reconnect_history[handle] = history

        assert all(re.search('^\d+$', h) for h in handles), "All handles must be numbers: {}".format(handles)
        allocation = self.with_retries(self._requestor.allocation_create,
            client_id=self.client_id,
            runtime_id=self.runtime_id,
            placement=self.placement,
            params=_params,
            handles=handles,
            initial=initial,
        )
        news = len([entry for entry in allocation['info']['n'] if entry['new']])
        extra_logger.info('[%s] Received allocation with %s new and %s existing envs: %s', self.label, news, len(allocation['info']['n']) - news, allocation)

        assert len(allocation['env_n']) <= len(handles), "Received more envs than requested: allocation={} handles={}".format(allocation, handles)
        _, pending = self._handle_allocation(allocation)

        for env in pending:
            self.pending[env['name']] = {
                'handle': env['handle'],
                'params': params,
                'received_at': time.time()
            }

    def _poll(self):
        self._sleep = min(20, self._sleep + 2)

        if len(self.pending) == 0:
            return

        for name, spec in self.pending.items():
            delta = time.time() - spec['received_at']
            if delta > self.start_timeout:
                raise error.TimeoutError('Waited {}s for {} to get an IP, which exceeds start_timeout of {}'.format(delta, name, self.start_timeout))

        names = list(self.pending.keys())
        # This really should be an allocation_get, but it's possible
        # the pods list will be long. So it's either GET with a body,
        # or POST what should really be a GET. We do the latter.
        allocation = self.with_retries(self._requestor.allocation_refresh, self.client_id, names=names)
        assert len(allocation['env_n']) <= len(names), "Received more envs than requested: allocation={} names={}".format(allocation, names)

        # Handle any envs which have gone missing
        result = set(env['name'] for env in allocation['env_n'])
        dropped = [p for p in self.pending.keys() if p not in result]
        if len(dropped) > 0:
            logger.info('Pending remote envs %s were not returned by the allocator (only %s were returned). Assuming the missing ones have gone down and requesting replacements.', dropped, list(result))
            for d in dropped:
                spec = self.pending.pop(d)
                self._allocate(dropped, False, spec['params'])

        # Handle successful allocations
        self._handle_allocation(allocation, pop=True)

    def _handle_allocation(self, allocation, pop=False):
        ready = []
        not_ready = []
        for alloc_env in allocation['env_n']:
            if alloc_env['status'] != 'allocated':
                not_ready.append(alloc_env)
                continue
            if pop:
                self.pending.pop(alloc_env['name'])
            vnc_address = alloc_env['vnc_recorder_address'] if self.use_recorder_ports else alloc_env['vnc_address']
            rewarder_address = alloc_env['rewarder_recorder_address'] if self.use_recorder_ports else alloc_env['rewarder_address']
            env = remote.Remote(
                name=alloc_env['name'],
                handle=alloc_env['handle'],
                vnc_address=vnc_address,
                vnc_password=alloc_env['vnc_password'],
                rewarder_address=rewarder_address,
                rewarder_password=alloc_env['rewarder_password'],
            )
            ready.append(env)

        if len(ready) > 0:
            extra_logger.info('[%s] The following envs now have IPs, but still may take time to boot: %s', self.label, ready)
            self.ready.put(ready)

        return ready, not_ready

    def with_retries(self, method, *args, **kwargs):
        timeout = 20 * 60
        start = time.time()

        i = 0
        while True:
            try:
                return method(*args, **kwargs)
            except FatalError as e:
                logger.error('[%s] %s', self.label, e)
                self.error_buffer.record(e)
                raise
            except Exception as e:
                delta = time.time() - start
                if delta > timeout:
                    raise error.TimeoutError('Have been unable to connect to the allocator at {} for {}s. Giving up. Last error: {}'.format(self.base_url, delta, e))
                i += 1

                sleep = min(2**i, 60)
                time.sleep(sleep)
                logger.error('[%s] Error making request to allocator: %s. Will retry in %ss (and timeout in %.0fs)', self.label, e, sleep, start + timeout - time.time())

class AllocatorClient(object):
    def __init__(self, label, api_key, base_url):
        self.label = label
        self.api_key = api_key
        self.base_url = base_url
        self.session = requests.Session()
        self.session.headers.update({'Content-type': 'application/json'})
        self.request_timeout = 80

    def _handle_resp(self, resp):
        try:
            parsed = resp.json()
        except ValueError as e:
            if resp.status_code == 500:
                raise RequestError(message="500 Internal Error (retrying automatically): response={}".format(resp.content), status_code=resp.status_code)
            elif resp.status_code == 503:
                raise RequestError(message="503 Error from server (allocator is probably overloaded): response={}".format(resp.content), status_code=resp.status_code)
            elif resp.status_code == 200:
                raise RequestError(message="Response from server: status_code={} response={}".format(resp.status_code, resp.content), status_code=resp.status_code)
            else:
                raise RequestError(message="Error from server: status_code={} response={}".format(resp.status_code, resp.content), status_code=resp.status_code)

        if resp.status_code == 200:
            return parsed
        elif 'detail' in parsed:
            raise FatalError(message=parsed['detail'], status_code=resp.status_code)
        else:
            raise RequestError(message='Malformed response from allocator, missing "detail" key: {}'.format(parsed), status_code=resp.status_code)

    def _post_request(self, route, data, description):
        url = urlparse.urljoin(self.base_url, route)
        extra_logger.info('[%s] %s: POST %s: %s', self.label, description, url, json.dumps(data))
        resp = self.session.post(urlparse.urljoin(self.base_url, route),
                                 data=json.dumps(data), auth=(self.api_key, ''),
                                 timeout=self.request_timeout,
        )
        return self._handle_resp(resp)

    def _delete_request(self, route):
        url = urlparse.urljoin(self.base_url, route)
        extra_logger.info("[%s] DELETE %s", self.label, url)
        resp = self.session.delete(url, auth=(self.api_key, ''), timeout=self.request_timeout)
        return self._handle_resp(resp)

    def _get_request(self, route):
        url = urlparse.urljoin(self.base_url, route)
        extra_logger.info("[%s] GET %s", self.label, url)
        resp = self.session.get(url, auth=(self.api_key, ''), timeout=self.request_timeout)
        return self._handle_resp(resp)

    def allocation_create(self, client_id, runtime_id, handles, params={}, placement='public', initial=False):
        route = '/v1/allocations'
        data = {'client': client_id, 'runtime': runtime_id, 'params': params, 'handles': handles, 'initial': initial, 'placement': placement}
        logger.info('Requesting %s environment%s from %s: %s', len(handles), 's' if len(handles) != 1 else '', self.base_url, data)
        resp = self._post_request(route, data, description='requesting new allocation')
        return resp

    def allocation_refresh(self, id, names):
        route = '/v1/allocations/{}'.format(id)
        resp = self._post_request(route, data={'names': names}, description='refreshing existing allocation')
        return resp

    def allocation_delete(self, id):
        route = '/v1/allocations/{}'.format(id)
        resp = self._post_request(route, {})
        return resp



import logging
import numpy as np
from universe import pyprofile
import threading
import time

from universe import error
from universe.twisty import reactor
from universe.vncdriver import server_messages
from universe.spaces import vnc_event

logger = logging.getLogger(__name__)

class NumpyScreen(object):
    def __init__(self, width, height):
        self.lock = threading.RLock()

        shape = (height, width, 3)
        self._screens = (np.zeros(shape, dtype=np.uint8), np.zeros(shape, dtype=np.uint8))

        self.color_cycle = [0, 1, 2]
        self._width = None
        self._height = None

        self._defer = []

        self.paint_cursor = False
        self._cursor = {
            id(self._screens[0]): {
                'behind': None,
                'details': None,
                'painted': False,
            },
            id(self._screens[1]): {
                'behind': None,
                'details': None,
                'painted': False,
            },
        }

        self._back_updated = True
        self._back_cursor_updated = True

        self.cursor_shape = None
        self.cursor_position = None

        self._has_initial_framebuffer_update = False

    def set_paint_cursor(self, paint_cursor):
        self.paint_cursor = paint_cursor

    def peek(self):
        front_screen, _ = self._screens
        return front_screen

    def flip(self):
        pyprofile.push('vncdriver.numpy_screen.flip_bitmap')
        with self.lock:
            if self._back_updated:
                updates = self._defer

                # Flip screens
                front_screen, back_screen = self._screens
                self._screens = back_screen, front_screen

                # Mark ourselves as pending application of updates
                self._back_updated = False

                # This can be called asynchronously if desired, but it means
                # less reliably smooth playback.
                #
                # reactor.callFromThread(self.update_back)
                self.update_back()
            else:
                updates = []
            result = self.peek(), {'vnc_session.framebuffer_updates': updates}
        pyprofile.pop()
        return result

    def apply_action(self, action):
        if isinstance(action, vnc_event.PointerEvent):
            with self.lock:
                self.cursor_position = (action.x, action.y)

                # If not self._back_updated, we're not actually up to
                # date, so any pixels we cached would be wrong. When
                # back updates, it'll automatically render the cursor.
                if self._back_updated and self.paint_cursor:
                    self._unpaint_cursor()
                    self._paint_cursor()

    def apply(self, framebuffer_update):
        with self.lock:
            self._has_initial_framebuffer_update = True

            # Pop any pending updates
            self._update_back()
            self._apply(framebuffer_update)
            self._defer.append(framebuffer_update)


    def _apply(self, framebuffer_update):
        if self.paint_cursor:
            self._unpaint_cursor()
        for rect in framebuffer_update.rectangles:
            if isinstance(rect.encoding,
                          (server_messages.RAWEncoding, server_messages.ZRLEEncoding, server_messages.ZlibEncoding)):
                self._update_rectangle(rect.x, rect.y, rect.width, rect.height, rect.encoding.data)
            elif isinstance(rect.encoding, server_messages.PseudoCursorEncoding):
                self._update_cursor_shape(rect.x, rect.y, rect.width, rect.height, rect.encoding.image, rect.encoding.mask)
            else:
                raise error.Error('Unrecognized encoding: {}'.format(rect.encoding))
        if self.paint_cursor:
            self._paint_cursor()

    def update_back(self):
        with self.lock:
            self._update_back()

    def _update_back(self):
        if self._back_updated:
            return
        self._back_updated = True

        for framebuffer_update in self._defer:
            self._apply(framebuffer_update)

        if len(self._defer) == 0 and self.paint_cursor:
            self._unpaint_cursor()
            self._paint_cursor()

        self._defer = []

    def _update_rectangle(self, x, y, width, height, data):
        _, back_screen = self._screens
        back_screen[y:y+height, x:x+width, self.color_cycle] = data

    def _update_cursor_shape(self, hotx, hoty, width, height, image, mask):
        # hotx, hoty are the hotspot within the cursor
        self.cursor_shape = (hotx, hoty, width, height, image, mask)

    def _paint_cursor(self):

        # use our knowledge of the x, y cursor position plus the
        # cursor shape to paint the cursor
        if self.cursor_position is None:
            return
        elif not self._has_initial_framebuffer_update:
            return
        elif self.cursor_shape is None:
            return

        self._back_cursor_updated = True

        _, back_screen = self._screens
        cursor = self._cursor[id(back_screen)]

        assert not cursor['painted']
        cursor['painted'] = True

        hotx, hoty, width, height, image, mask = self.cursor_shape
        x, y = self.cursor_position

        # Save old data
        cursor['details'] = (x, y, width, height)
        cursor['behind'] = back_screen[y:y+height, x:x+width, :].copy()

        # Paint the cursor
        total_h, total_w, _ = back_screen.shape

        cutoff_h = min(total_h - y, height)
        cutoff_w = min(total_w - x, width)
        image = image[:cutoff_h, :cutoff_w]
        mask = mask[:cutoff_h, :cutoff_w, np.newaxis]

        back_screen[y:y+height, x:x+width, self.color_cycle] = (1 - mask)*back_screen[y:y+height, x:x+width, self.color_cycle] + mask*image

    def _unpaint_cursor(self):

        _, back_screen = self._screens
        cursor = self._cursor[id(back_screen)]

        if cursor['behind'] is not None:
            assert cursor['painted']
            x, y, width, height = cursor['details']
            back_screen[y:y+height, x:x+width, :] = cursor['behind']
        cursor['painted'] = False


    # def _copy_rectangle(self, screen, src_x, src_y, x, y, width, height):
    #     update = np.frombuffer(data, dtype=np.uint8)
    #     update = update.reshape([height, width, 4])
    #     update = update[:, :, self._color_cycle]  # Ignore X channel


    #     screen[y:y+height, x:x+width] = screen[src_y:src_y+height, src_x:src_x+width]

    # def _fill_rectangle(self, screen, x, y, width, height, color):
    #     update = np.frombuffer(color, dtype=np.uint8)
    #     update = update[self._color_cycle]
    #     screen[y:y+height, x:x+width] = update

    # def begin(self, number_of_rectangles):
    #     self.lock.acquire()
    #     # This may already have been called via
    #     # reactor.callFromThread. It's safe to be called multiple times.
    #     self._update_back()

    # def commit(self):
    #     self.lock.release()

    # def back_bitmap(self):
    #     _, back_screen = self._screens
    #     return back_screen

    # def screen_synced(self):
    #     # TODO: lock?
    #     return self._screens is not None



import logging
import time
from universe import pyprofile, rewarder, spaces, vectorized

logger = logging.getLogger(__name__)

class Throttle(vectorized.Wrapper):
    """
    A env wrapper that makes sending the action ASAP.

    Previous implementation would sleep first and then call env._step.
    This implementation calls env._step twice:
        1. first call submits given action
        2. after sleeping based on fps, second call submits empty action to
           receive observation.

    visual observation from first call is discarded.
    metadata and rewards from the two calls are merged.
    text observations are merged as well.
    """
    def __init__(self, env):
        super(Throttle, self).__init__(env)

        self._steps = None

    def _configure(self, skip_metadata=False, fps='default', **kwargs):
        super(Throttle, self)._configure(**kwargs)
        if fps == 'default':
            fps = self.metadata['video.frames_per_second']
        self.fps = fps
        self.skip_metadata = skip_metadata

        self.diagnostics = self.unwrapped.diagnostics

    def _reset(self):
        # We avoid aggregating reward/info across episode boundaries
        # by caching it on the object
        self._deferred_reward_n = None
        self._deferred_done_n = None
        self._deferred_info_n = None

        observation = self.env.reset()
        self._start_timer()
        return observation

    def _step(self, action_n):
        if self._steps is None:
            self._start_timer()
        self._steps += 1

        accum_observation_n, accum_reward_n, accum_done_n, accum_info = self._substep(action_n)
        accum_info['throttle.action.available_at'] = time.time()

        # Record which indexes we were just peeking at, so when we
        # make the follow-up we'll be sure to peek there too.
        peek_n = [any(spaces.PeekReward for peek in action) for action in action_n]

        if self.fps is None:
            return accum_observation_n, accum_reward_n, accum_done_n, accum_info

        accum_info['stats.throttle.sleep'] = 0
        while True:
            # See how much time we have to idle
            delta = self._start + 1./self.fps * self._steps - time.time()

            # The following assumes that our control loop
            if delta < 0:
                # We're out of time. Just get out of here.
                delta = abs(delta)
                if delta >= 1:
                    logger.info('Throttle fell behind by %.2fs; lost %.2f frames', delta, self.fps*delta)
                pyprofile.timing('vnc_env.Throttle.lost_sleep', delta)
                self._start_timer()
                break
            # elif delta < 0.008:
            #     # Only have 8ms. Let's spend it sleeping, and
            #     # return an image which may have up to an
            #     # additional 8ms lag.
            #     #
            #     # 8ms is reasonably arbitrary; we just want something
            #     # that's small where it's not actually going to help
            #     # if we make another step call. Step with 32 parallel
            #     # envs takes about 6ms (about half of which is
            #     # diagnostics, which could be totally async!), so 8 is
            #     # a reasonable choice for now..
            #     pyprofile.timing('vnc_env.Throttle.sleep', delta)
            #     accum_info['stats.throttle.sleep'] += delta
            #     time.sleep(delta)
            #     break
            else:
                # We've got plenty of time. Sleep for up to 16ms, and
                # then refresh our current frame. We need to
                # constantly be calling step so that our lags are
                # reported correctly, within 16ms. (The layering is
                # such that the vncdriver doesn't know which pixels
                # correspond to metadata, and the diagnostics don't
                # know when pixels first got painted. So we do our
                # best to present frames as they're ready to the
                # diagnostics.)
                delta = min(delta, 0.016)
                pyprofile.timing('vnc_env.Throttle.sleep', delta)
                accum_info['stats.throttle.sleep'] += delta
                time.sleep(delta)

                # We want to merge in the latest reward/done/info so that our
                # agent has the most up-to-date info post-sleep, but also want
                # to avoid popping any rewards where done=True (since we'd
                # have to merge across episode boundaries).
                action_n = []
                for done, peek in zip(accum_done_n, peek_n):
                    if done or peek:
                        # No popping of reward/done
                        action_n.append([spaces.PeekReward])
                    else:
                        action_n.append([])

                observation_n, reward_n, done_n, info = self._substep(action_n)

                # Merge observation, rewards and metadata.
                # Text observation has order in which the messages are sent.
                rewarder.merge_n(
                    accum_observation_n, accum_reward_n, accum_done_n, accum_info,
                    observation_n, reward_n, done_n, info,
                )

        return accum_observation_n, accum_reward_n, accum_done_n, accum_info

    def _substep(self, action_n):
        with pyprofile.push('vnc_env.Throttle.step'):
            start = time.time()
            # Submit the action ASAP, before the thread goes to sleep.
            observation_n, reward_n, done_n, info = self.env.step(action_n)

            available_at = info['throttle.observation.available_at'] = time.time()
            if available_at - start > 1:
                logger.info('env.step took a long time: %.2fs', available_at - start)
            if not self.skip_metadata and self.diagnostics is not None:
                # Run (slow) diagnostics
                self.diagnostics.add_metadata(observation_n, info['n'], available_at=available_at)
            return observation_n, reward_n, done_n, info

    def _start_timer(self):
        self._start = time.time()
        self._steps = 0



from universe.envs import vnc_env
from universe.spaces import VNCActionSpace


class WorldOfGooEnv(vnc_env.VNCEnv):
    def __init__(self):
        super(WorldOfGooEnv, self).__init__()
        # TODO: set action space screen shape to match
        # HACK: empty keys list fails for some weird reason, give it an 'a'
        self.action_space = VNCActionSpace(keys=['a'], buttonmasks=[1])



class JoystickEvent(object):
    pass


class JoystickAxisEvent(JoystickEvent):
    def __init__(self, amount):
        self.amount = float(amount)

    def __repr__(self):
        return str(type(self)) + '<amount={}>'.format(self.amount)

    def __str__(self):
        return repr(self)

    def __hash__(self):
        return self.amount.__hash__()

    def __eq__(self, other):
        return type(other) == type(self) and \
               other.amount == self.amount

    def compile(self):
        return type(self).__name__, self.amount


class JoystickAxisXEvent(JoystickAxisEvent):
    pass


class JoystickAxisYEvent(JoystickAxisEvent):
    pass


class JoystickAxisZEvent(JoystickAxisEvent):
    pass


class JoystickAxisRxEvent(JoystickAxisEvent):
    pass


class JoystickAxisRyEvent(JoystickAxisEvent):
    pass


class JoystickAxisRzEvent(JoystickAxisEvent):
    pass


class JoystickSlider0Event(JoystickAxisEvent):
    pass


class JoystickSlider1Event(JoystickAxisEvent):
    pass
